from pylab import *
import unittest

from rockit import Ocp, DirectMethod, MultipleShooting, FreeTime, Stage, external_method
import numpy as np
from casadi import kron, DM
import casadi as ca
from rockit.casadi_helpers import AutoBrancher

class AcadosTests(unittest.TestCase):

    def test_time_dependence(self):
      for ab in AutoBrancher():

        t0_free = ab.branch() # False
        T_free = ab.branch() # True

        ocp = Ocp(t0=FreeTime(0) if t0_free else 0.03,T=FreeTime(1.0) if T_free else 1.5)

        if t0_free:
          ocp.subject_to(ocp.t0 == 0.03)

        p = ocp.state()
        v = ocp.state()

        u = ocp.control()

        ocp.set_der(p,v)
        if ab.branch():
          ocp.set_der(v,u+ocp.t)
        else:
          ocp.set_der(v,2*u) # here

        ocp.add_objective(ocp.T)
        #ocp.add_objective(ocp.integral(v**2+u**2))
        if not T_free:
          ocp.add_objective(ocp.integral(u**2))
        ocp.add_objective(ocp.at_tf(v-1)**2)

        case = ab.branch(range(6))
        if case==0:
          ocp.subject_to(-1<= (u <= 1))
        elif case==1:
          ocp.subject_to(-1<= (1.2*u <= 1),include_last=False)
        elif case==2:
          ocp.subject_to(-1.3<= (u-ocp.t <= 1),include_last=False)
        elif case==3:
          ocp.subject_to(-1.3<= (u <= 1+ocp.t),include_last=False)
        elif case==4:
          ocp.subject_to(-1.3-ocp.t <= (u<= 1.1+ocp.t),include_last=False)
        elif case==5:
          ocp.subject_to(-1<= u)
          ocp.subject_to(u <= 1+ocp.t,include_last=False)

        ocp.subject_to(ocp.at_t0(p)==0)
        ocp.subject_to(ocp.at_t0(v)==0)

        ocp.subject_to(ocp.at_tf(p)==1)

        ocp.solver('ipopt',{"ipopt.tol":1e-10})

        ocp.set_initial(u,1)
        ocp.set_initial(p, ocp.t)

        # Obtain reference solution
        N = 5
        ocp.method(MultipleShooting(N=N))
        sol = ocp.solve()

        signals = [("p",p),("v",v),("u",u),("t",ocp.t)]
        values = [("T",ocp.T)]#,("T1",ocp.at_t0(ocp.t)),("T2",ocp.at_tf(ocp.t))]

        ref = {}
        for k, expr in signals:
          ref[k] = sol.sample(expr,grid='control')[1]
        for k, expr in values:
          ref[k] = sol.value(expr)

        [ref_t,_] = sol.sample(signals[0][1],grid='control')

        for method in [external_method('acados', N=N,qp_solver='PARTIAL_CONDENSING_HPIPM',nlp_solver_max_iter=200,hessian_approx='EXACT',regularize_method = 'CONVEXIFY',integrator_type='ERK',nlp_solver_type='SQP',qp_solver_cond_N=N,tol=1e-8)]:




          ocp.method(method)
          sol = ocp.solve()

          print("analysis here")
          print(ocp.T)
          print(sol.value(ocp.T))

          sold = {}
          for k, expr in signals:
            sold[k] = sol.sample(expr,grid='control')[1]
          for k, expr in values:
            sold[k] = sol.value(expr)

          [sol_t,_] = sol.sample(signals[0][1],grid='control')

          tolerance = 1e-6

          for k, expr in signals:
            print(k+"ref",ref[k])
            print(k+"sol",sold[k])
          for k, expr in values:
            print(k+"ref",ref[k])
            print(k+"sol",sold[k])
          print("ref_t",ref_t)
          print("sol_t",sol_t)

          for k, expr in signals:
            np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
          for k, expr in values:
            np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
          np.testing.assert_allclose(sol_t, ref_t, atol=tolerance)


    def test_modes(self):
      for ab in AutoBrancher():

        ocp = Ocp(T=1.0)

        x      = ocp.state()
        v      = ocp.state()

        F       = ocp.control()

        ocp.set_der(x, v)
        ocp.set_der(v, F)

        if ab.branch():
          ocp.add_objective(ocp.sum(v**2))
        else:
          ocp.add_objective(ocp.integral(v**2))
        ocp.add_objective(5*ocp.at_tf(v)**2)

        ocp.subject_to(ocp.at_t0(x)==0)
        if ab.branch():
          ocp.subject_to(ocp.at_t0(v)==0.0)
        else:
          pass

        ocp.subject_to(ocp.at_tf(x)==5.0)
        ocp.subject_to(F<=15.0)
        ocp.subject_to(v<=7.0,include_first=ab.branch(),include_last=ab.branch())
      
        ocp.method(MultipleShooting(N=4,intg='rk'))

        ocp.solver('ipopt',{"ipopt.tol":1e-10})

        sol = ocp.solve()

        signals = [("x",x),("v",v),("F",F)]
        values = []#("mayer",mayer)]#[("slack",slack)]

        ref = {}
        for k, expr in signals:
          ref[k] = sol.sample(expr,grid='control')[1]
        for k, expr in values:
          ref[k] = sol.value(expr)

        [ref_t,_] = sol.sample(signals[0][1],grid='control')

        print(ref)


        for method in [external_method('acados',N=4,qp_solver='PARTIAL_CONDENSING_HPIPM',nlp_solver_max_iter=2000,hessian_approx='EXACT',regularize_method = 'CONVEXIFY',integrator_type='ERK',nlp_solver_type='SQP',qp_solver_cond_N=4,tol=1e-8)]:

          ocp.method(method)
          sol = ocp.solve()

          sold = {}
          for k, expr in signals:
            sold[k] = sol.sample(expr,grid='control')[1]
          for k, expr in values:
            sold[k] = sol.value(expr)

          [sol_t,_] = sol.sample(signals[0][1],grid='control')

          tolerance = 1e-5

          for k, expr in signals:
            print(k+"ref",ref[k])
            print(k+"sol",sold[k])
          for k, expr in values:
            print(k+"ref",ref[k])
            print(k+"sol",sold[k])
          print("ref_t",ref_t)
          print("sol_t",sol_t)

          for k, expr in signals:
            np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
          for k, expr in values:
            np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
          np.testing.assert_allclose(sol_t, ref_t, atol=tolerance)

    def test_slacks(self):
      
      obj_store = None

      for ab in AutoBrancher():
        mode = ab.branch(["nominal","perturbed","soft"])

        ocp = Ocp(T=1.0)

        x      = ocp.state()
        v      = ocp.state()

        F       = ocp.control()

        ocp.set_der(x, v)
        ocp.set_der(v, F)

        ocp.add_objective(ocp.sum(v**2))
        ocp.add_objective(5*ocp.at_tf(v)**2)

        ocp.subject_to(ocp.at_t0(x)==0.0)
        ocp.subject_to(ocp.at_t0(v)==0.0)

        constraints = []

        
        constr_type = ab.branch(["simple","linear","nonlinear"])
        if constr_type=="simple":
          constraints.append((lambda e: F <= 15+e,True,{})) # simple
          constraints.append((lambda e: v <= 7+e,True,dict(include_first=False,include_last=False)))  # include_first=True induces failure
          constraints.append((lambda e: ocp.at_tf(x) + e>=5.0,False,{})) # boundary constraint on state
        elif constr_type=="linear":
          constraints.append((lambda e: F <= 15+e,True,{})) # simple
          constraints.append((lambda e: v+0.01*x <= 7+e,True,dict(include_first=True,include_last=False)))  # include_first=True induces failure
          constraints.append((lambda e: ocp.at_tf(x+0.01*v) + e>=5.0,False,{})) # boundary constraint on state
        elif constr_type=="nonlinear":
          constraints.append((lambda e: F <= 15+e,True,{})) # simple
          constraints.append((lambda e: v+0.01*x**2 <= 7+e,True,dict(include_first=False,include_last=False)))  # include_first=True induces failure
          constraints.append((lambda e: ocp.at_tf(x+0.01*v**2) + e>=5.0,False,{}))



        if mode=="nominal":
          for c,_,kwargs in constraints:
            ocp.subject_to(c(0),**kwargs)
        elif mode=="perturbed":
          k = ab.branch(range(len(constraints)))
          for i,(c,_,kwargs) in enumerate(constraints):
            ocp.subject_to(c(0.1 if i==k else 0),**kwargs)
        elif mode=="soft":
          k = ab.branch(range(len(constraints)))
          power = ab.branch([1,2]) # 1-norm or 2-norm?
          for i,(c,signal,kwargs) in enumerate(constraints):
            if i!=k:
              ocp.subject_to(c(0),**kwargs)
            else:
              slack = ocp.variable(grid='control' if signal else '')
              ocp.subject_to(slack>=0)
              ocp.subject_to(c(slack),**kwargs)
              penalty = 0.3*slack**power
              print("signal",signal, penalty)
              ocp.add_objective(ocp.sum(penalty) if signal else penalty)

        ocp.method(MultipleShooting(N=4,intg='rk'))

        ocp.solver('ipopt',{"ipopt.tol":1e-10})

        sol = ocp.solve()

        # Are the constraints we added really active?
        if mode=="nominal":
          obj_store = sol.value(ocp.objective)
        else:
          print("constraint",k)
          with self.assertRaises(Exception):
            np.testing.assert_allclose(sol.value(ocp.objective), obj_store, atol=0.01)

        signals = [("x",x),("v",v),("F",F)]
        values = []#("mayer",mayer)]#[("slack",slack)]

        ref = {}
        for k, expr in signals:
          ref[k] = sol.sample(expr,grid='control')[1]
        for k, expr in values:
          ref[k] = sol.value(expr)

        [ref_t,_] = sol.sample(signals[0][1],grid='control')

        print(ref)

        #raise Exception("")

        for method in [external_method('acados',N=4,qp_solver='PARTIAL_CONDENSING_HPIPM',nlp_solver_max_iter=2000,hessian_approx='EXACT',regularize_method = 'CONVEXIFY',integrator_type='ERK',nlp_solver_type='SQP',qp_solver_cond_N=4)]:

          ocp.method(method)
          sol = ocp.solve()

          sold = {}
          for k, expr in signals:
            sold[k] = sol.sample(expr,grid='control')[1]
          for k, expr in values:
            sold[k] = sol.value(expr)

          [sol_t,_] = sol.sample(signals[0][1],grid='control')

          tolerance = 1e-5

          for k, expr in signals:
            print(k+"ref",ref[k])
            print(k+"sol",sold[k])
          for k, expr in values:
            print(k+"ref",ref[k])
            print(k+"sol",sold[k])
          print("ref_t",ref_t)
          print("sol_t",sol_t)

          for k, expr in signals:
            np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
          for k, expr in values:
            np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
          np.testing.assert_allclose(sol_t, ref_t, atol=tolerance)

    def test_parameters(self):
        
      def reference(p_global_value,p_local_value):
        ocp = Ocp(t0=0, T=10)

        # Define two scalar states (vectors and matrices also supported)
        x1 = ocp.state()
        x2 = ocp.state()

        # Define one piecewise constant control input
        #  (use `order=1` for piecewise linear)
        u = ocp.control()

        # Compose time-dependent expressions a.k.a. signals
        #  (explicit time-dependence is supported with `ocp.t`)
        e = 1 - x2**2

        # Specify differential equations for states
        #  (DAEs also supported with `ocp.algebraic` and `add_alg`)
        ocp.set_der(x1, e * x1 - x2 + p_global_value*u)
        ocp.set_der(x2, x1)

        # Lagrange objective term: signals in an integrand
        ocp.add_objective(ocp.sum(x1**2 + (x2-p_local_value(ocp.t))**2 + u**2))
        # Mayer objective term: signals evaluated at t_f = t0_+T
        ocp.add_objective(ocp.at_tf(x1**2))

        # Path constraints
        #  (must be valid on the whole time domain running from `t0` to `tf`,
        #   grid options available such as `grid='inf'`)
        ocp.subject_to(x1 >= -0.25)
        ocp.subject_to(-1 <= (u <= 1 ))

        # Boundary constraints
        ocp.subject_to(ocp.at_t0(x1) == 0)
        ocp.subject_to(ocp.at_t0(x2) == 1)

        # Pick an NLP solver backend
        #  (CasADi `nlpsol` plugin):
        ocp.solver('ipopt')

        # Pick a solution method
        method = external_method('acados', N=10,qp_solver='PARTIAL_CONDENSING_HPIPM',nlp_solver_max_iter=200,hessian_approx='EXACT',regularize_method = 'CONVEXIFY',integrator_type='ERK',nlp_solver_type='SQP',qp_solver_cond_N=10)
        ocp.method(method)

        # Set initial guesses for states, controls and variables.
        #  Default: zero
        ocp.set_initial(x2, 0)                 # Constant
        ocp.set_initial(x1, ocp.t/10)          # Function of time
        ocp.set_initial(u, linspace(0, 0.1, 10)) # Matrix

        # Solve
        sol = ocp.solve()

        return sol.sample(ocp.x,grid='control')[1]
      
      ref = reference(1,lambda t : 1)
      ref_var_global = reference(1.1,lambda t : 1)
      ref_var_local = reference(1,lambda t : 1.1)
      ref_var_local2 = reference(1,lambda t : 1+0.01*t)
      ref_var = reference(1.1,lambda t : 1+0.01*t)

      ocp = Ocp(t0=0, T=10)

      # Define two scalar states (vectors and matrices also supported)
      x1 = ocp.state()
      x2 = ocp.state()

      # Define one piecewise constant control input
      #  (use `order=1` for piecewise linear)
      u = ocp.control()

      p_local = ocp.parameter(grid='control')
      ocp.set_value(p_local,[1+0*i*0.01 for i in range(11)])

      p_global = ocp.parameter()
      ocp.set_value(p_global,1)

      # Compose time-dependent expressions a.k.a. signals
      #  (explicit time-dependence is supported with `ocp.t`)
      e = 1 - x2**2

      # Specify differential equations for states
      #  (DAEs also supported with `ocp.algebraic` and `add_alg`)
      ocp.set_der(x1, e * x1 - x2 + p_global*u)
      ocp.set_der(x2, x1)

      # Lagrange objective term: signals in an integrand
      ocp.add_objective(ocp.sum(x1**2 + (x2-p_local)**2 + u**2))
      # Mayer objective term: signals evaluated at t_f = t0_+T
      ocp.add_objective(ocp.at_tf(x1**2))

      # Path constraints
      #  (must be valid on the whole time domain running from `t0` to `tf`,
      #   grid options available such as `grid='inf'`)
      ocp.subject_to(x1 >= -0.25)
      ocp.subject_to(-1 <= (u <= 1 ))

      # Boundary constraints
      ocp.subject_to(ocp.at_t0(x1) == 0)
      ocp.subject_to(ocp.at_t0(x2) == 1)

      # Pick an NLP solver backend
      #  (CasADi `nlpsol` plugin):
      ocp.solver('ipopt')

      # Pick a solution method
      method = external_method('acados', N=10,qp_solver='PARTIAL_CONDENSING_HPIPM',nlp_solver_max_iter=200,hessian_approx='EXACT',regularize_method = 'CONVEXIFY',integrator_type='ERK',nlp_solver_type='SQP',qp_solver_cond_N=10)
      ocp.method(method)

      # Set initial guesses for states, controls and variables.
      #  Default: zero
      ocp.set_initial(x2, 0)                 # Constant
      ocp.set_initial(x1, ocp.t/10)          # Function of time
      ocp.set_initial(u, linspace(0, 0.1, 10)) # Matrix
    
      # Solve
      sol = ocp.solve()

      res = sol.sample(ocp.x,grid='control')[1]
      print(res)

      f = ocp.to_function('f',[ocp.sample(ocp.x,grid='control')[1], ocp.sample(u,grid='control-')[1] ],[ocp.sample(ocp.x,grid='control')[1]])

      np.testing.assert_allclose(res, np.array(f(0,0)).T, atol=1e-5)

      np.testing.assert_allclose(ref, np.array(f(0,0)).T, atol=1e-5)

      f = ocp.to_function('f',[p_global, ocp.sample(p_local,grid='control')[1], ocp.sample(ocp.x,grid='control')[1], ocp.sample(u,grid='control-')[1] ],[ocp.sample(ocp.x,grid='control')[1]])

      np.testing.assert_allclose(ref, np.array(f(1,1,0,0)).T, atol=1e-5)
      np.testing.assert_allclose(ref_var_global, np.array(f(1.1,1,0,0)).T, atol=1e-5)
      np.testing.assert_allclose(ref_var_local, np.array(f(1,1.1,0,0)).T, atol=1e-5)
      np.testing.assert_allclose(ref_var_local2, np.array(f(1,ca.hcat([1+i*0.01 for i in range(11)]),0,0)).T, atol=1e-5)

      np.testing.assert_allclose(ref_var, np.array(f(1.1,ca.hcat([1+i*0.01 for i in range(11)]),0,0)).T, atol=1e-5)

if __name__ == '__main__':
    unittest.main()



# -*- coding: future_fstrings -*-
#
# Copyright 2019 Gianluca Frison, Dimitris Kouzoupis, Robin Verschueren,
# Andrea Zanelli, Niels van Duijkeren, Jonathan Frey, Tommaso Sartor,
# Branimir Novoselnik, Rien Quirynen, Rezart Qelibari, Dang Doan,
# Jonas Koenemann, Yutao Chen, Tobias Schöls, Jonas Schlagenhauf, Moritz Diehl
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

import os
import sys

from .utils import verbose_system_call


class CMakeBuilder:
    """
    Class to work with the `CMake` build system.
    """
    def __init__(self):
        self._source_dir = None  # private source directory, this is set to code_export_dir
        self.build_dir = 'build'
        self._build_dir = None  # private build directory, usually rendered to abspath(build_dir)
        self.generator = None
        self.host = None
        """Defines the generator, options can be found via `cmake --help` under 'Generator'. Type: string. Linux default 'Unix Makefiles', Windows 'Visual Studio 15 2017 Win64'; default value: `None`."""
        # set something for Windows
        if os.name == 'nt':
            self.generator = 'Visual Studio 15 2017 Win64'
        self.build_targets = None
        """A comma-separated list of the build targets, if `None` then all targets will be build; type: List of strings; default: `None`."""
        self.options_on = None
        """List of strings as CMake options which are translated to '-D Opt[0]=ON -D Opt[1]=ON ...'; default: `None`."""

    # Generate the command string for handling the cmake command.
    def get_cmd1_cmake(self):
        defines_str = ''
        if self.options_on is not None:
            defines_arr = [f' -D{opt}=ON' for opt in self.options_on]
            defines_str = ' '.join(defines_arr)
        generator_str = ''
        if self.generator is not None:
            generator_str = f' -G"{self.generator}"'
        host_str = ''
        if self.host is not None:
            host_str = f' -A"{self.host}"'
        return f'cmake -DCMAKE_BUILD_TYPE=Release -DCMAKE_INSTALL_PREFIX="{self._source_dir}"{defines_str}{generator_str}{host_str} -Wdev -S"{self._source_dir}" -B"{self._build_dir}"'

    # Generate the command string for handling the build.
    def get_cmd2_build(self):
        import multiprocessing
        cmd = f'cmake --build "{self._build_dir}" --config Release -j{multiprocessing.cpu_count()}'
        if self.build_targets is not None:
            cmd += f' -t {self.build_targets}'
        return cmd

    # Generate the command string for handling the install command.
    def get_cmd3_install(self):
        return f'cmake --install "{self._build_dir}"'

    def exec(self, code_export_directory, verbose=True):
        """
        Execute the compilation using `CMake` with the given settings.
        :param code_export_directory: must be the absolute path to the directory where the code was exported to
        """
        if(os.path.isabs(code_export_directory) is False):
            print(f'(W) the code export directory "{code_export_directory}" is not an absolute path!')
        self._source_dir = code_export_directory
        self._build_dir = os.path.abspath(self.build_dir)
        try:
            os.mkdir(self._build_dir)
        except FileExistsError as e:
            pass

        try:
            os.chdir(self._build_dir)
            cmd_str = self.get_cmd1_cmake()
            print(f'call("{cmd_str})"')
            retcode = verbose_system_call(cmd_str, verbose, shell=True)
            if retcode != 0:
                raise RuntimeError(f'CMake command "{cmd_str}" was terminated by signal {retcode}')
            cmd_str = self.get_cmd2_build()
            print(f'call("{cmd_str}")')
            retcode = verbose_system_call(cmd_str, verbose, shell=True)
            if retcode != 0:
                raise RuntimeError(f'Build command "{cmd_str}" was terminated by signal {retcode}')
            cmd_str = self.get_cmd3_install()
            print(f'call("{cmd_str}")')
            retcode = verbose_system_call(cmd_str, verbose, shell=True)
            if retcode != 0:
                raise RuntimeError(f'Install command "{cmd_str}" was terminated by signal {retcode}')
        except OSError as e:
            print("Execution failed:", e, file=sys.stderr)
        except Exception as e:
            print("Execution failed:", e, file=sys.stderr)
            exit(1)



def ocp_get_default_cmake_builder() -> CMakeBuilder:
    """
    If :py:class:`~acados_template.acados_ocp_solver.AcadosOcpSolver` is used with `CMake` this function returns a good first setting.
    :return: default :py:class:`~acados_template.builders.CMakeBuilder`
    """
    cmake_builder = CMakeBuilder()
    cmake_builder.options_on = ['BUILD_ACADOS_OCP_SOLVER_LIB']
    return cmake_builder


def sim_get_default_cmake_builder() -> CMakeBuilder:
    """
    If :py:class:`~acados_template.acados_sim_solver.AcadosSimSolver` is used with `CMake` this function returns a good first setting.
    :return: default :py:class:`~acados_template.builders.CMakeBuilder`
    """
    cmake_builder = CMakeBuilder()
    cmake_builder.options_on = ['BUILD_ACADOS_SIM_SOLVER_LIB']
    return cmake_builder



from rockit import *
import numpy as np

ocp = Ocp(T=1.0)

x      = ocp.state()
v      = ocp.state()

F       = ocp.control()

ocp.set_der(x, v)
ocp.set_der(v, F)

ocp.add_objective(ocp.sum(v**2))
ocp.add_objective(5*ocp.at_tf(v)**2)

ocp.subject_to(ocp.at_t0(x)==0)
ocp.subject_to(ocp.at_t0(v)==3.03030303)

ocp.subject_to(ocp.at_tf(x)==5.0)
ocp.subject_to(F<=15.0)
ocp.subject_to(v<=7.0,include_first=False,include_last=False)

ocp.method(MultipleShooting(N=4,intg='rk'))

ocp.solver('ipopt',{"ipopt.tol":1e-10})


#ocp.set_initial(x, [-1.46569578e-13,  7.77175578e-01 , 2.27513837e+00 , 4.02513837e+00, 5.00000000e+00])
#ocp.set_initial(v, [1.23370231, 4.98370231, 7.   ,      7.     ,    0.79889307])
#ocp.set_initial(F, [ 1.50000000e+01 , 8.06519075e+00 ,-2.23072479e-10, -2.48044277e+01])

sol = ocp.solve()

signals = [("x",x),("v",v),("F",F)]
values = []#("mayer",mayer)]#[("slack",slack)]

ref = {}
for k, expr in signals:
  ref[k] = sol.sample(expr,grid='control')[1]
for k, expr in values:
  ref[k] = sol.value(expr)

[ref_t,_] = sol.sample(signals[0][1],grid='control')

print(ref)


for method in [external_method('acados',N=4,qp_solver='PARTIAL_CONDENSING_HPIPM',nlp_solver_max_iter=2000,hessian_approx='EXACT',regularize_method = 'CONVEXIFY',integrator_type='ERK',nlp_solver_type='SQP',qp_solver_cond_N=4,tol=1e-8)]:

  ocp.method(method)
  sol = ocp.solve()

  sold = {}
  for k, expr in signals:
    sold[k] = sol.sample(expr,grid='control')[1]
  for k, expr in values:
    sold[k] = sol.value(expr)

  [sol_t,_] = sol.sample(signals[0][1],grid='control')

  tolerance = 1e-5

  for k, expr in signals:
    print(k+"ref",ref[k])
    print(k+"sol",sold[k])
  for k, expr in values:
    print(k+"ref",ref[k])
    print(k+"sol",sold[k])
  print("ref_t",ref_t)
  print("sol_t",sol_t)

  for k, expr in signals:
    np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
  for k, expr in values:
    np.testing.assert_allclose(sold[k], ref[k], atol=tolerance)
  np.testing.assert_allclose(sol_t, ref_t, atol=tolerance)



# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;

from dataclasses import dataclass, field
import numpy as np


@dataclass
class ZoroDescription:
    """
    Zero-Order Robust Optimization (zoRO) scheme.

    The uncertainty propagation is performed by:
    $$P_{k+1} = (A_k + B_kK)P_k(A_k + B_kK)^\top + GWG^\top$$

    For advanced users.
    """
    backoff_scaling_gamma: float = 1.0
    """backoff scaling factor, for stochastic MPC"""
    fdbk_K_mat: np.ndarray = None
    """constant feedback gain matrix K"""
    unc_jac_G_mat: np.ndarray = None    # default: an identity matrix
    """matrix G, describes how noise enters the dynamics"""
    P0_mat: np.ndarray = None
    """initial uncertainty matrix $\bar{P}_0$"""
    W_mat: np.ndarray = None
    """matrix W, covariance of noise in stochastic setting, defines uncertainty ellipsoids in robust setting"""
    idx_lbx_t: list = field(default_factory=list)
    """Indices of constraints to be tightened within the lower bounds on x for intermediate shooting nodes 1,...,N-1"""
    idx_ubx_t: list = field(default_factory=list)
    """Indices of constraints to be tightened within the upper bounds on x for intermediate shooting nodes 1,...,N-1"""
    idx_lbx_e_t: list = field(default_factory=list)
    """Indices of constraints to be tightened within the lower bounds on x for terminal shooting node"""
    idx_ubx_e_t: list = field(default_factory=list)
    """Indices of constraints to be tightened within the upper bounds on x for terminal shooting node"""
    idx_lbu_t: list = field(default_factory=list)
    """Indices of constraints to be tightened within the lower bounds on u for intermediate shooting nodes 1,...,N-1"""
    idx_ubu_t: list = field(default_factory=list)
    """Indices of constraints to be tightened within the upper bounds on u for intermediate shooting nodes 1,...,N-1"""
    idx_lg_t: list = field(default_factory=list)
    idx_ug_t: list = field(default_factory=list)
    idx_lg_e_t: list = field(default_factory=list)
    idx_ug_e_t: list = field(default_factory=list)
    idx_lh_t: list = field(default_factory=list)
    idx_uh_t: list = field(default_factory=list)
    idx_lh_e_t: list = field(default_factory=list)
    idx_uh_e_t: list = field(default_factory=list)
    # Inputs:
    input_P0_diag: bool = False
    """Determines if diag(P0) is an input to the custom update function"""
    input_P0: bool = True
    """Determines if P0 is an input to the custom update function, specified in column-major format"""

    input_W_diag: bool = False
    """Determines if diag(W) is an input to the custom update function"""
    input_W_add_diag: bool = False
    """
    Determines if the concatenation of diag(W_{add}^k) is an input to the custom update function

    In case this is used W_k = W + W_{add}^k.
    """

    # Outputs:
    output_P_matrices: bool = False
    """Determines if the matrices P_k are outputs of the custom update function"""


def process_zoro_description(zoro_description: ZoroDescription):
    zoro_description.nw, _ = zoro_description.W_mat.shape
    if zoro_description.unc_jac_G_mat is None:
        zoro_description.unc_jac_G_mat = np.eye(zoro_description.nw)
    zoro_description.nlbx_t = len(zoro_description.idx_lbx_t)
    zoro_description.nubx_t = len(zoro_description.idx_ubx_t)
    zoro_description.nlbx_e_t = len(zoro_description.idx_lbx_e_t)
    zoro_description.nubx_e_t = len(zoro_description.idx_ubx_e_t)
    zoro_description.nlbu_t = len(zoro_description.idx_lbu_t)
    zoro_description.nubu_t = len(zoro_description.idx_ubu_t)
    zoro_description.nlg_t = len(zoro_description.idx_lg_t)
    zoro_description.nug_t = len(zoro_description.idx_ug_t)
    zoro_description.nlg_e_t = len(zoro_description.idx_lg_e_t)
    zoro_description.nug_e_t = len(zoro_description.idx_ug_e_t)
    zoro_description.nlh_t = len(zoro_description.idx_lh_t)
    zoro_description.nuh_t = len(zoro_description.idx_uh_t)
    zoro_description.nlh_e_t = len(zoro_description.idx_lh_e_t)
    zoro_description.nuh_e_t = len(zoro_description.idx_uh_e_t)

    if zoro_description.input_P0_diag and zoro_description.input_P0:
        raise Exception("Only one of input_P0_diag and input_P0 can be True")

    # Print input note:
    print(f"\nThe data of the generated custom update function consists of the concatenation of:")
    i_component = 1
    if zoro_description.input_P0_diag:
        print(f"{i_component}) input: diag(P0)")
        i_component += 1
    if zoro_description.input_P0:
        print(f"{i_component}) input: P0; full matrix in column-major format")
        i_component += 1
    if zoro_description.input_W_diag:
        print(f"{i_component}) input: diag(W)")
        i_component += 1
    if zoro_description.input_W_add_diag:
        print(f"{i_component}) input: concatenation of diag(W_gp^k) for i=0,...,N-1")
        i_component += 1
    if zoro_description.output_P_matrices:
        print(f"{i_component}) output: concatenation of colmaj(P^k) for i=0,...,N")
        i_component += 1
    print("\n")

    return zoro_description



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from casadi import integrator, Function, MX, hcat, vertcat, vcat, linspace, veccat, DM, repmat, horzsplit, cumsum, inf, mtimes, symvar, horzcat, symvar, vvcat, is_equal
import casadi as ca
from .direct_method import DirectMethod
from .splines import BSplineBasis, BSpline
from .casadi_helpers import reinterpret_expr, HashOrderedDict, HashDict, is_numeric
from numpy import nan, inf
import numpy as np
from collections import defaultdict
from .splines.micro_spline import bspline_derivative, eval_on_knots, get_greville_points
from .casadi_helpers import get_ranges_dict

# Agnostic about free or fixed start or end point: 

# Agnostic about time?
class Grid:
    def __init__(self, min=0, max=inf):
        self.min = min
        self.max = max

    def __call__(self, t0, T, N):
        return linspace(t0+0.0, t0+T+0.0, N+1)

    def bounds_finalize(self, opti, control_grid, t0_local, tf, N):
        pass

class BSplineSignal:
    def __init__(self, coeff, xi, degree,T=1,parametric=False):
        """
        Parameters
        ----------
        coeff : :obj:`casadi.MX`
            Coefficients of the spline
        xi : :obj:`casadi.MX`
            Grid of the spline (no duplicates)
        degree : int
            Degree of the spline
        T : float, optional
            Time scaling of the spline
            Default: 1
        parametric : bool, optional
            Whether the spline is ony dependant on parameters
        """
        self.coeff = coeff
        self.xi = xi
        self.degree = degree
        [_,self.B] = eval_on_knots(xi, degree)
        self.sampled = horzsplit(coeff @ self.B)
        self.derivative = None
        self.derivative_of = None
        self.T = T
        self.parametric = parametric

        # Initialization delegated to register
        self.peers = None
        self.symbol = None
    
    def sample(self,**kwargs):
        [_,B] = eval_on_knots(self.xi, self.degree, **kwargs)
        return self.coeff @ B

    @property
    def der(self):
        if self.derivative is not None:
            if self.degree==0:
                raise Exception("Cannot differentiate " + self.symbol.name() + " any further.")
            der_symbol = MX.sym("der_"+self.symbol.name(), self.symbol.sparsity())
            self.derivative = self.get_der()
            self.derivative.parametric = self.parametric
            self.derivative.derivative_of = self
            self.peers[der_symbol] = self.derivative
        return self.derivative.symbol

    def get_der(self):
        return BSplineSignal(bspline_derivative(self.coeff,self.xi,self.degree)/self.T, self.xi, self.degree-1, T=self.T)

    @staticmethod
    def register(peers, symbol, stage, signal):
        peers[symbol] = signal
        signal.peers = peers
        signal.symbol = symbol

        # Register derivative if present in stage._signals
        target = stage._signals[symbol]
        if target.derivative is not None:
            signal_der = signal.get_der()
            signal.derivative = signal_der
            signal_der.parametric = signal.parametric
            signal_der.derivative_of = signal
            BSplineSignal.register(peers, target.derivative.symbol, stage, signal_der)

class FixedGrid(Grid):
    def __init__(self, localize_t0=False, localize_T=False, **kwargs):
        Grid.__init__(self, **kwargs)
        self.localize_t0 = localize_t0
        self.localize_T = localize_T

    def bounds_T(self, T_local, t0_local, k, T, N):
        if self.localize_T:
            Tk = T_local[k]
            if k>=0 and k+1<N:
                r = self.constrain_T(T_local[k], T_local[k+1], N)
                if r is not None:
                    yield r
        else:
            Tk = T*(self.normalized(N)[k+1]-self.normalized(N)[k])
        if self.localize_t0 and k>=0:
            yield (t0_local[k]+Tk==t0_local[k+1],{})

    def get_t0_local(self, opti, k, t0, N):
        if self.localize_t0:
            if k==0:
                return t0
            else:
                return opti.variable()
        else:
            return None

    def get_T_local(self, opti, k, T, N):
        if self.localize_T:
            if k==0:
                return T*self.scale_first(N)
            else:
                return opti.variable()
        else:
            return None

class FreeGrid(FixedGrid):
    """Specify a grid with unprescribed spacing
    
    Parameters
    ----------
    min : float or  :obj:`casadi.MX`, optional
        Minimum size of control interval
        Enforced with constraints
        Default: 0
    max : float or  :obj:`casadi.MX`, optional
        Maximum size of control interval
        Enforced with constraints
        Default: inf
    """
    def __init__(self, **kwargs):
        FixedGrid.__init__(self, **kwargs)
        self.localize_T = True

    def constrain_T(self,T,Tnext,N):
        pass

    def bounds_T(self, T_local, t0_local, k, T, N):
        yield (self.min <= (T_local[k] <= self.max),{})
        for i,e in enumerate(FixedGrid.bounds_T(self, T_local, t0_local, k, T, N)):
            yield e

    def bounds_finalize(self, opti, control_grid, t0_local, tf, N):
        opti.subject_to(control_grid[-1]==tf)

    def get_T_local(self, opti, k, T, N):
        return opti.variable()

class UniformGrid(FixedGrid):
    """Specify a grid with uniform spacing
    
    Parameters
    ----------
    min : float or  :obj:`casadi.MX`, optional
        Minimum size of control interval
        Enforced with constraints
        Default: 0
    max : float or  :obj:`casadi.MX`, optional
        Maximum size of control interval
        Enforced with constraints
        Default: inf
    """
    def __init__(self, **kwargs):
        FixedGrid.__init__(self, **kwargs)

    def constrain_T(self,T,Tnext,N):
        return (Tnext==T,{})

    def scale_first(self, N):
        return 1.0/N

    def bounds_T(self, T_local, t0_local, k, T, N):

        for i,e in enumerate(FixedGrid.bounds_T(self, T_local, t0_local, k, T, N)):
            yield e
            if i==0 and k==0:
                if self.localize_T:
                    yield (self.min <= (T_local[0] <= self.max), {})
                else:
                    if self.min==0 and self.max==inf:
                        pass
                    else:
                        yield (self.min <= (T/N <= self.max), {})

    def normalized(self, N):
        return list(np.linspace(0.0, 1.0, N+1))


class FunctionGrid(FixedGrid):
    def __init__(self, normalized_fun, **kwargs):
        """
        Inputs:
            normalized: function that takes N and returns a list of N+1 normalized grid points

        """
        self.normalized_fun = normalized_fun
        FixedGrid.__init__(self, **kwargs)

    def __call__(self, t0, T, N):
        n = self.normalized(N)
        return t0 + hcat(n)*T

    def normalized(self, N):
        return self.normalized_fun(N)
class DensityGrid(FixedGrid):
    def __init__(self, density, integrator='cvodes',integrator_options=None,**kwargs):
        """
        Expression in one symbolic variable (dimensionless time) that describes the density of the grid

        e.g. t**2

        We first compute the definite integral of the density over the interval [0,1]:        
        I = integral_0^t density dt

        
        Next, we inspect the function
        
        E(t) := 1/I*integral_0^t density dt
        
        The grid points t_i are the computed such that E(t_i) is a uniform parition of [0,1]

        """
        self.density = density
        self.t = symvar(density)[0]

        self.integrator_options = {} if integrator_options is None else integrator_options
        self.integrator = integrator
        self.cache = {}

        FixedGrid.__init__(self, **kwargs)

    def __call__(self, t0, T, N):
        n = self.normalized(N)
        return t0 + hcat(n)*T

    def normalized(self, N):
        if N in self.cache: return self.cache[N]
        import scipy
        import scipy.optimize
        x = MX.sym("x")
        scale = MX.sym("scale")
        ode = {"x": x, "ode": scale*ca.substitute(self.density,self.t,self.t*scale), "p": scale, "t":self.t}
        intg = integrator('intg', self.integrator, ode, 0, 1, self.integrator_options)
        I = float(intg(x0=0,p=1)["xf"])
        res = [0]
        for v in list(np.linspace(0.0, 1.0, N+1)[1:-1]*I):
            r = scipy.optimize.root_scalar(lambda tau: float(intg(x0=0,p=tau)["xf"]-v), method='bisect', bracket=[0,1])
            res.append(r.root)
        res.append(1.0)
        self.cache[N] = res
        return res
    
class DenseEdgesGrid(DensityGrid):
    def __init__(self, multiplier=10, edge_frac=0.1, **kwargs):
        interp = ca.interpolant('interp','bspline',[[0.0,edge_frac,1-edge_frac,1.0]],[multiplier,1.0,1.0,multiplier],{"algorithm":"smooth_linear"})
        tau = ca.MX.sym("tau")
        DensityGrid.__init__(self, interp(tau), **kwargs)

class GeometricGrid(FixedGrid):
    """Specify a geometrically growing grid
    
    Each control interval is a constant factor larger than its predecessor

    Parameters
    ----------
    growth_factor : float>1
        See `local` for interpretation
    local : bool, optional
        if False, last interval is growth_factor larger than first
        if True, interval k+1 is growth_factor larger than interval k
        Default: False
    min : float or  :obj:`casadi.MX`, optional
        Minimum size of control interval
        Enforced with constraints
        Default: 0
    max : float or  :obj:`casadi.MX`, optional
        Maximum size of control interval
        Enforced with constraints
        Default: inf

    Examples
    --------

    >>> MultipleShooting(N=3, grid=GeometricGrid(2)) # grid = [0, 1, 3, 7]*T/7
    >>> MultipleShooting(N=3, grid=GeometricGrid(2,local=True)) # grid = [0, 1, 5, 21]*T/21
    """
    def __init__(self, growth_factor, local=False, **kwargs):
        assert growth_factor>=1
        self._growth_factor = growth_factor
        self.local = local
        FixedGrid.__init__(self, **kwargs)

    def __call__(self, t0, T, N):
        n = self.normalized(N)
        return t0 + hcat(n)*T

    def growth_factor(self, N):
        if not self.local and N>1:
            return self._growth_factor**(1.0/(N-1))
        return self._growth_factor

    def constrain_T(self,T,Tnext,N):
        return (T*self.growth_factor(N)==Tnext,{})

    def scale_first(self, N):
        return self.normalized(N)[1]

    def bounds_T(self, T_local, t0_local, k, T, N):
        if self.localize_T:
            if k==0 or k==-1:
                yield (self.min <= (T_local[k] <= self.max), {})
        else:
            n = self.normalized(N)
            if k==0:
                yield (self.min <= (T*n[1] <= self.max),{})
            if k==-1:
                yield (self.min <= (T*(n[-1]-n[-2]) <= self.max),{})
        for e in FixedGrid.bounds_T(self, T_local, t0_local, k, T, N):
            yield e

    def normalized(self, N):
        g = self.growth_factor(N)
        base = 1.0
        vec = [0]
        for i in range(N):
            vec.append(vec[-1]+base)
            base *= g
        for i in range(N+1):
            vec[i] = vec[i]/vec[-1]
        return vec

class SamplingMethod(DirectMethod):
    def __init__(self, N=50, M=1, intg='rk', intg_options=None, grid=UniformGrid(), **kwargs):
        """
        Parameters
        ----------
        grid : `str` or tuple of :obj:`casadi.MX`
            List of arbitrary expression containing states, controls, ...
        name : `str`
            Name for CasADi Function
        options : dict, optional
            Options for CasADi Function
        intg : `str`
            Rockit-specific methods: 'rk', 'expl_euler' - these allow to make use of signal sampling with 'refine'
            Others are passed on as-is to CasADi
        """
        DirectMethod.__init__(self, **kwargs)
        self.N = N
        self.M = M
        self.intg = intg
        self.intg_options = {} if intg_options is None else intg_options
        self.time_grid = grid
        self.clean()

    def clean(self):
        self.X = []  # List that will hold N+1 decision variables for state vector
        self.Q = []  # List that will hold N+1 expressions for quadrature states
        self.U = []  # List that will hold N decision variables for control vector
        self.Z = []  # Algebraic vars

        self.T = None # Will be set to numbers/opti.variables
        self.t0 = None

        self.P = []
        self.V = None
        self.V_control = []
        self.V_control_plus = []
        self.V_states = []
        self.P_control = []
        self.P_control_plus = []
        self.signals = HashOrderedDict()

        self.poly_coeff = []  # Optional list to save the coefficients for a polynomial
        self.poly_coeff_z = []  # Optional list to save the coefficients for a polynomial
        self.poly_coeff_q = []  # Optional list to save the coefficients for a polynomial
        self.xk = []  # List for intermediate integrator states
        self.xqk = []
        self.zk = []
        self.xr = []
        self.zr = []
        self.tr = []
        self.q = 0

    def discrete_system(self, stage):
        # Coefficient matrix from RK4 to reconstruct 4th order polynomial (k1,k2,k3,k4)
        # nstates x (4 * M)
        poly_coeffs_c = []
        poly_coeffs_z = []
        poly_coeffs_c_q = []
        poly_coeffs_d = []
        poly_coeffs_d_q = []

        t0 = MX.sym('t0')
        T = MX.sym('T')
        DT = T / self.M

        # Size of integrator interval
        X0 = MX.sym("x", stage.nx)            # Initial state
        U = MX.sym("u", stage.nu)             # Control
        P = MX.sym("p", stage.np+stage.v.shape[0])
        Z = MX.sym("z", stage.nz)

        Z0 = MX.sym("Z0", stage.nz)


        Zs = []

        # Compute local start time
        t0_local = t0

        Qc = []
        Qd = []

        ode, diffeq, partition, composition, qpartition, qcomposition = stage._hybrid()

        XQ_dummy = MX.zeros(stage.nxq)
        qc0, qd0 = partition(XQ_dummy)

        quad_c = DM.zeros(qc0.shape[0])
        quad_d = DM.zeros(qd0.shape[1])
        Xc0, Xd0 = partition(X0)


        Xc0_sym = MX.sym('Xc0', Xc0.shape)
        Xd0_sym = MX.sym('Xd0', Xd0.shape)


        if hasattr(self, 'intg_' + self.intg):
            intg = getattr(self, "intg_" + self.intg)(ode, Xc0_sym, Xd0_sym, U, P, Z)
        else:
            intg = self.intg_builtin(ode, Xc0_sym, Xd0_sym, U, P, Z)

        Xc = [Xc0]
        Xd = [Xd0]

        Z0_current = Z0

        for j in range(self.M):
            intg_res = intg(x_c0=Xc[-1], x_d=Xd[-1], u=U, t0=t0_local, DT=DT, DT_control=T, p=P, z0=Z0_current)
            diffeq_res = diffeq(x_c=Xc[-1], x_d=Xd[-1], u=U, p=P, t0=t0_local, DT=DT, DT_control=T)
            Xc.append(intg_res["xf"])
            Xd.append(diffeq_res["xf"])
            Zs.append(intg_res["zf"])
            poly_coeffs_c.append(intg_res["poly_coeff"])
            poly_coeffs_z.append(intg_res["poly_coeff_z"])
            poly_coeffs_c_q.append(intg_res["poly_coeff_q"])
            # If the integrator for continuous states does not support polynomial output
            if poly_coeffs_c[0].shape==(0,0):
                # Forget about the polynomial output for discrete states too
                poly_coeffs_d.append(MX())
            else:
                # Create a fixed-value polynomial for discrete states
                poly_coeffs_d.append(horzcat(diffeq_res["xf"], DM.zeros(diffeq_res["xf"].shape[0],poly_coeffs_c[0].shape[1]-1)))
            if poly_coeffs_d[0].shape==(0,0):
                poly_coeffs_d_q.append(MX())
            else:
                poly_coeffs_d_q.append(horzcat(diffeq_res["qf"], DM.zeros(diffeq_res["qf"].shape[0],poly_coeffs_c_q[0].shape[1]-1)))

            quad_c = quad_c + intg_res["qf"]
            quad_d = quad_d + diffeq_res["qf"]
            Qc.append(quad_c)
            Qd.append(quad_d)

            t0_local += DT
            Z0_current = intg_res["zf"]

        Qi = qcomposition(hcat(Qc),hcat(Qd))
        Xi = composition(hcat(Xc),hcat(Xd))

        poly_coeffs = composition(hcat(poly_coeffs_c),hcat(poly_coeffs_d))
        poly_coeffs_q = qcomposition(hcat(poly_coeffs_c_q),hcat(poly_coeffs_d_q))

        ret = Function('F', [X0, U, T, t0, P, Z0], [
                        composition(Xc[-1],Xd[-1]),
                        Xi,
                        poly_coeffs,
                        qcomposition(quad_c,quad_d),
                        Qi,
                        poly_coeffs_q,
                        Zs[-1],
                        hcat(Zs),
                        hcat(poly_coeffs_z)],
                       ['x0', 'u', 'T', 't0', 'p', 'z0'], ['xf', 'Xi', 'poly_coeff', 'qf', 'Qi', 'poly_coeff_q', 'zf', 'Zi', 'poly_coeff_z'])
        assert not ret.has_free()
        return ret

    def intg_rk(self, f, X_c, X_d, U, P, Z):
        assert Z.is_empty()
        DT = MX.sym("DT")
        DT_control = MX.sym("DT_control")
        t0 = MX.sym("t0")
        Z0 = MX.sym("z0", 0, 1)
        # A single Runge-Kutta 4 step
        k1 = f(x_c=X_c, x_d=X_d, u=U, p=P, t=t0)
        k2 = f(x_c=X_c + DT / 2 * k1["ode"], x_d=X_d, u=U, p=P, t=t0+DT/2)
        k3 = f(x_c=X_c + DT / 2 * k2["ode"], x_d=X_d, u=U, p=P, t=t0+DT/2)
        k4 = f(x_c=X_c + DT * k3["ode"], x_d=X_d, u=U, p=P, t=t0+DT)

        f0 = k1["ode"]
        f1 = 2/DT*(k2["ode"]-k1["ode"])/2
        f2 = 4/DT**2*(k3["ode"]-k2["ode"])/6
        f3 = 4*(k4["ode"]-2*k3["ode"]+k1["ode"])/DT**3/24
        poly_coeff = hcat([X_c, f0, f1, f2, f3])

        f0 = k1["quad"]
        f1 = 2/DT*(k2["quad"]-k1["quad"])/2
        f2 = 4/DT**2*(k3["quad"]-k2["quad"])/6
        f3 = 4*(k4["quad"]-2*k3["quad"]+k1["quad"])/DT**3/24
        poly_coeff_q = hcat([f0, f1, f2, f3])

        return Function('F', [X_c, X_d, U, t0, DT, DT_control, P, Z0], [X_c + DT / 6 * (k1["ode"] + 2 * k2["ode"] + 2 * k3["ode"] + k4["ode"]), poly_coeff, DT / 6 * (k1["quad"] + 2 * k2["quad"] + 2 * k3["quad"] + k4["quad"]), poly_coeff_q, MX(0, 1), MX()], ['x_c0', 'x_d', 'u', 't0', 'DT', 'DT_control', 'p', 'z0'], ['xf', 'poly_coeff', 'qf', 'poly_coeff_q', 'zf', 'poly_coeff_z'])

    def intg_expl_euler(self, f, X_c, X_d, U, P, Z):
        assert Z.is_empty()
        DT = MX.sym("DT")
        DT_control = MX.sym("DT_control")
        t0 = MX.sym("t0")
        Z0 = MX.sym("z0", 0, 1)
        k = f(x_c=X_c, x_d=X_d, u=U, p=P, t=t0)
        poly_coeff = hcat([X_c, k["ode"]])
        poly_coeff_q = k["quad"]

        return Function('F', [X_c, X_d, U, t0, DT, DT_control, P, Z0], [X_c + DT * k["ode"], poly_coeff, DT * k["quad"], poly_coeff_q, MX(0, 1), MX()], ['x_c0', 'x_d', 'u', 't0', 'DT', 'DT_control', 'p', 'z0'], ['xf', 'poly_coeff', 'qf', 'poly_coeff_q', 'zf', 'poly_coeff_z'])

    def intg_builtin(self, f, X_c, X_d, U, P, Z):
        # A single CVODES step
        DT = MX.sym("DT")
        DT_control = MX.sym("DT_control")
        t = MX.sym("t")
        t0 = MX.sym("t0")
        Z0 = MX.sym("Z0", Z.sparsity())
        res = f(x_c=X_c, x_d=X_d, u=U, p=P, t=t0+t*DT, z=Z)
        data = {'x': X_c, 'p': vertcat(U, DT, DT_control, P, t0, X_d), 'z': Z, 't': t, 'ode': DT * res["ode"], 'quad': DT * res["quad"], 'alg': res["alg"]}
        options = dict(self.intg_options)
        if self.intg in ["collocation"]:
            # In rockit, M replaces number_of_finite_elements on a higher level
            if "number_of_finite_elements" not in options:
                options["number_of_finite_elements"] = 1
        I = integrator('intg_'+self.intg, self.intg, data, options)
        if I.size2_out("xf")!=1:
            raise Exception("Integrator must only return outputs at a single timepoint. Did you specify a grid?")
        res = I.call({'x0': X_c, 'p': vertcat(U, DT, DT_control, P, t0, X_d), 'z0': Z0})
        return Function('F', [X_c, X_d, U, t0, DT, DT_control, P, Z0], [res["xf"], MX(), res["qf"], MX(), res["zf"], MX()], ['x_c0', 'x_d', 'u', 't0', 'DT', 'DT_control', 'p', 'z0'], ['xf', 'poly_coeff','qf','poly_coeff_q','zf','poly_coeff_z'])

    def untranscribe_placeholders(self, phase, stage):
        pass

    def transcribe_placeholders(self, phase, stage, placeholders):
        """
        Transcription is the process of going from a continuous-time OCP to an NLP
        """
        return stage._transcribe_placeholders(phase, self, placeholders)

    def transcribe_start(self, stage, opti):
        return
        
    def untranscribe(self, stage, phase=1,**kwargs):
        self.clean()

    def transcribe_event_after_varpar(self, stage, phase=1, **kwargs):
        pass

    def transcribe(self, stage, phase=1,**kwargs):
        """
        Transcription is the process of going from a continuous-time OCP to an NLP
        """
        if phase==0: return
        opti = stage.master._method.opti
        if phase==1:

            DM.set_precision(14)

            self.transcribe_start(stage, opti)

            # Grid for B-spline
            self.xi = ca.vec(DM(self.time_grid(0, 1, self.N))).T

            # Parameters needed before variables because of self.T = self.eval(stage, stage._T)
            self.add_parameter(stage, opti)
            self.add_variables(stage, opti)
            self.add_parameter_signals(stage, opti)
            self.set_parameter(stage, opti)

            self.transcribe_event_after_varpar(stage, phase=phase, **kwargs)

            self.integrator_grid = []
            for k in range(self.N):
                t_local = linspace(self.control_grid[k], self.control_grid[k+1], self.M+1)
                self.integrator_grid.append(t_local[:-1] if k<self.N-1 else t_local)
            #self.add_constraints_before(stage, opti)
            self.add_constraints(stage, opti)
            self.add_constraints_after(stage, opti)
            self.add_objective(stage, opti)
        if phase==2:

            self.set_initial(stage, opti, stage._initial)
            T_init = opti.debug.value(self.T, opti.initial())
            t0_init = opti.debug.value(self.t0, opti.initial())

            initial = HashOrderedDict()
            # How to get initial value -> ask opti?
            control_grid_init = self.time_grid(t0_init, T_init, self.N)
            if self.time_grid.localize_t0:
                for k in range(1, self.N):
                    initial[self.t0_local[k]] = control_grid_init[k]
                initial[self.t0_local[self.N]] = control_grid_init[self.N]
            if self.time_grid.localize_T:
                for k in range(not isinstance(self.time_grid, FreeGrid), self.N):
                    initial[self.T_local[k]] = control_grid_init[k+1]-control_grid_init[k]

            self.set_initial(stage, opti, initial)
            self.set_initial(stage, opti, stage._initial) # Redo this: ocp.t is correct only now
            self.set_parameter(stage, opti)


    def add_constraints_before(self, stage, opti):
        for c, meta, args in stage._constraints["point"]:
            e = self.eval(stage, c)
            if 'r_at_tf' not in [a.name() for a in symvar(e)]:
                opti.subject_to(e, args["scale"], meta=meta)

    def add_constraints_after(self, stage, opti):
        for c, meta, args in stage._constraints["point"]:
            e = self.eval(stage, c)
            if 'r_at_tf' in [a.name() for a in symvar(e)]:
                opti.subject_to(e, args["scale"], meta=meta)

    def add_inf_constraints(self, stage, opti, c, k, l, meta):
        # Query the discretization method used for polynomial coefficients
        #   interpretation: state ~= coeff * [t^0;t^1;t^2;...]
        #                    t is physical time, but starting at 0 at the beginning of the interval
        coeff = stage._method.poly_coeff[k * self.M + l]

        # Represent polynomial as a BSpline object (https://gitlab.kuleuven.be/meco-software/rockit/-/blob/v0.1.28/rockit/splines/spline.py#L392)
        degree = coeff.shape[1]-1
        basis = BSplineBasis([0]*(degree+1)+[1]*(degree+1),degree)
        tscale = self.T / self.N / self.M
        tpower = vcat([tscale**i for i in range(degree+1)])
        coeff = coeff * repmat(tpower.T,stage.nx,1)
        # TODO: bernstein transformation as function of degree
        Poly_to_Bernstein_matrix_4 = DM([[1,0,0,0,0],[1,1.0/4, 0, 0, 0],[1, 1.0/2, 1.0/6, 0, 0],[1, 3.0/4, 1.0/2, 1.0/4, 0],[1, 1, 1, 1, 1]])
        # Use a direct way to obtain Bernstein coefficients from polynomial coefficients
        state_coeff = mtimes(Poly_to_Bernstein_matrix_4,coeff.T)
        
        # Replace symbols for states by BSpline object derivatives
        subst_from = list(stage.states)

        statesize = [0] + [elem.nnz() for elem in stage.states]
        statessizecum = np.cumsum(statesize)
        state_coeff_split = horzsplit(state_coeff,statessizecum)
        subst_to = [BSpline(basis,coeff) for coeff in state_coeff_split]


        lookup = dict(zip(subst_from, subst_to))

        # Replace symbols for state derivatives by BSpline object derivatives
        subst_from += stage._inf_der.keys()
        dt = (self.control_grid[k + 1] - self.control_grid[k])/self.M
        subst_to += [lookup[e].derivative()*(1/dt) for e in stage._inf_der.values()]

        subst_from += stage._inf_inert.keys()
        subst_to += stage._inf_inert.values()

        # Here, the actual replacement takes place
        c_spline = reinterpret_expr(c, subst_from, subst_to)

        # The use of '>=' or '<=' on BSpline objects automatically results in
        # those same operations being relayed onto BSpline coefficients
        # see https://gitlab.kuleuven.be/meco-software/rockit/-/blob/v0.1.28/rockit/splines/spline.py#L520-521
        try:
            opti.subject_to(self.eval_at_control(stage, c_spline, k), meta=meta)
        except IndexError:
            pass
    def fill_placeholders_integral_control(self, phase, stage, expr, refine=1):
        if phase==1: return
        [ts,exprs] = stage._sample(expr,grid='control',refine=refine)
        return ca.sum2(ca.diff(ts).T*exprs[:,:-1])
        r = 0
        for k in range(self.N):
            dt = self.control_grid[k + 1] - self.control_grid[k]
            r = r + self.eval_at_control(stage, expr, k)*dt
        return r

    def fill_placeholders_sum_control(self, phase, stage, expr, *args):
        if phase==1: return
        r = 0
        for k in range(self.N):
            r = r + self.eval_at_control(stage, expr, k)
        return r

    def fill_placeholders_sum_control_plus(self, phase, stage, expr, *args):
        if phase==1: return
        r = 0
        for k in list(range(self.N))+[-1]:
            r = r + self.eval_at_control(stage, expr, k)
        return r

    def placeholders_next(self, stage, expr, *args):
        self.eval_at_control(stage, expr, k+1)

    def fill_placeholders_at_t0(self, phase, stage, expr, *args):
        if phase==1: return
        return self.eval_at_control(stage, expr, 0)

    def fill_placeholders_at_tf(self, phase, stage, expr, *args):
        if phase==1: return
        return self.eval_at_control(stage, expr, -1)

    def add_objective(self, stage, opti):
        opti.add_objective(self.eval(stage, stage._objective))

    def add_variables_V(self, stage, opti):
        DirectMethod.add_variables(self, stage, opti)

        # Create time grid (might be symbolic)
        self.T = self.eval(stage, stage._T)
        self.t0 = self.eval(stage, stage._t0)

        self.t0_local = [None]*(self.N+1)
        self.T_local = [None]*self.N

        for p in stage.variables['bspline']:
            cat = stage._catalog[p]
            # Compute the degree and size of a BSpline coefficient needed
            d = cat["order"]
            s = self.N+d
            assert p.size2()==1
            C = opti.variable(p.size1(), s)
            BSplineSignal.register(self.signals, p, stage, BSplineSignal(C, self.xi, d, T=self.T))


    def add_variables_V_control(self, stage, opti, k):
        if k==0:
            self.V_control = [[] for v in stage.variables['control']]
            self.V_control_plus = [[] for v in stage.variables['control+']]
            for i, v in enumerate(stage.variables['states']):
                self.V_states.append([opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v], domain=stage._catalog[v]['domain'])])
        for i, v in enumerate(stage.variables['control']):
            self.V_control[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v], domain=stage._catalog[v]['domain']))
        for i, v in enumerate(stage.variables['control+']):
            self.V_control_plus[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v], domain=stage._catalog[v]['domain']))
        for i, v in enumerate(stage.variables['states']):
            self.V_states[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v], domain=stage._catalog[v]['domain']))

        self.t0_local[k] = self.time_grid.get_t0_local(opti, k, self.t0, self.N)
        self.T_local[k] = self.time_grid.get_T_local(opti, k, self.T, self.N)

    def add_variables_V_control_finalize(self, stage, opti):
        for i, v in enumerate(stage.variables['control+']):
            self.V_control_plus[i].append(opti.variable(v.shape[0], v.shape[1], scale=stage._scale[v], domain=stage._catalog[v]['domain']))
        if self.time_grid.localize_t0:
            self.t0_local[self.N] = opti.variable()
            self.control_grid = hcat(self.t0_local)

        elif self.time_grid.localize_T:
            t0 = self.t0
            cumsum = [t0]
            for e in self.T_local:
                cumsum.append(cumsum[-1]+e)
            self.control_grid = hcat(cumsum)
        else:
            self.control_grid = self.time_grid(self.t0, self.T, self.N)

        self.time_grid.bounds_finalize(opti, self.control_grid, self.t0_local, self.t0+self.T, self.N)

    def add_coupling_constraints(self, stage, opti, k):
        advanced = opti.advanced
        for c,kwargs in self.time_grid.bounds_T(self.T_local, self.t0_local, k, self.T, self.N):
            if not advanced.is_parametric(c):
                opti.subject_to(c,**kwargs)

    def get_p_control_at(self, stage, k=-1):
        return veccat(*[p[k] for p in self.P_control])

    def get_v_control_at(self, stage, k=-1):
        return veccat(*[v[k] for v in self.V_control])

    def get_p_control_plus_at(self, stage, k=-1):
        return veccat(*[p[k] for p in self.P_control_plus])

    def get_v_control_plus_at(self, stage, k=-1):
        return veccat(*[v[k] for v in self.V_control_plus])

    def get_v_states_at(self, stage, k=-1):
        return veccat(*[v[k] for v in self.V_states])

    def get_signals_at(self, stage, k=-1):
        return veccat(*[e.sampled[k] for e in self.signals.values()])

    def get_p_sys(self, stage, k, include_signals=True):
        args = [vvcat(self.P),
                self.get_p_control_at(stage, k),
                self.get_p_control_plus_at(stage, k),
                self.V, self.get_v_control_at(stage, k),
                self.get_v_control_plus_at(stage, k)]
        if include_signals:
            args.append(self.get_signals_at(stage, k))
        return vcat(args)

    def eval(self, stage, expr):
        return stage.master._method.eval_top(stage.master,
                                             stage._expr_apply(expr,
                                                               p=veccat(*self.P),
                                                               v=self.V,
                                                               t0=stage.t0,
                                                               T=stage.T))

    def eval_at_control(self, stage, expr, k):
        try:
            syms = symvar(expr)
        except:
            syms = []
        offsets = defaultdict(list)
        symbols = defaultdict(list)
        for s in syms:
            if s in stage._offsets:
                e, offset = stage._offsets[s]
                offsets[offset].append(e)
                symbols[offset].append(s)

        subst_from = []
        subst_to = []
        for offset in offsets.keys():
            if k==-1 and offset>0:
                raise IndexError()
            if k+offset<0:
                raise IndexError()
            subst_from.append(vvcat(symbols[offset]))
            subst_to.append(self._eval_at_control(stage, vvcat(offsets[offset]), k+offset))
            #print(expr, subst_from, subst_to)


        DT_control = self.get_DT_control_at(k)
        DT = self.get_DT_at(k, self.M-1 if k==-1 else 0)
        if self.Q:
            xq = self.Q[k]
        else:
            if k==-1:
                xq = self.q
            else:
                xq = nan
        expr = stage._expr_apply(expr,
                                 sub=(subst_from, subst_to),
                                 t0=self.t0,
                                 T=self.T,
                                 x=self.X[k],
                                 z=self.Z[k] if self.Z else nan,
                                 xq=xq,
                                 u=self.U[k],
                                 p_control=self.get_p_control_at(stage, k),
                                 p_control_plus=self.get_p_control_plus_at(stage, k),
                                 v=self.V, p=veccat(*self.P),
                                 v_control=self.get_v_control_at(stage, k),
                                 v_control_plus=self.get_v_control_plus_at(stage, k),
                                 signals=(self.signals, self.get_signals_at(stage, k)),
                                 v_states=self.get_v_states_at(stage, k),
                                 t=self.control_grid[k],
                                 DT=DT,
                                 DT_control=DT_control)
        expr = stage.master._method.eval_top(stage.master, expr)
        return expr
    
    def get_DT_control_at(self, k):
        if k==-1 or k==self.N:
            return self.control_grid[-1]-self.control_grid[-2]
        return self.control_grid[k+1]-self.control_grid[k]

    def get_DT_at(self, k, i):
        integrator_grid = self.integrator_grid[k]
        if i<integrator_grid.numel()-1:
            return integrator_grid[i+1]-integrator_grid[i]
        else:
            return self.integrator_grid[k+1][0]-self.integrator_grid[k][i]

    def _eval_at_control(self, stage, expr, k):
        x = self.X[k]
        if self.Q:
            xq = self.Q[k]
        else:
            if k==-1:
                xq = self.q
            else:
                xq = nan
        z = self.Z[k] if self.Z else nan
        u = self.U[-1] if k==len(self.U) else self.U[k] # Would be fixed if we remove the (k=-1 case)
        p_control = self.get_p_control_at(stage, k) if k!=len(self.U) else self.get_p_control_at(stage, k-1)
        v_control = self.get_v_control_at(stage, k) if k!=len(self.U) else self.get_v_control_at(stage, k-1)
        p_control_plus = self.get_p_control_plus_at(stage, k)
        v_control_plus = self.get_v_control_plus_at(stage, k)
        v_states = self.get_v_states_at(stage, k)
        t = self.control_grid[k]
        DT_control = self.get_DT_control_at(k)
        if k==-1 or k==len(self.integrator_grid):
            DT = self.get_DT_at(len(self.integrator_grid)-1, self.M-1)
        else:
            DT = self.get_DT_at(k, 0)
        return stage._expr_apply(expr,
                                 t0=self.t0,
                                 T=self.T,
                                 x=x,
                                 z=z,
                                 xq=xq,
                                 u=u,
                                 p_control=p_control,
                                 p_control_plus=p_control_plus,
                                 v=self.V, p=veccat(*self.P),
                                 v_control=v_control,
                                 v_control_plus=v_control_plus,
                                 v_states=v_states,
                                 t=t,
                                 DT=DT,
                                 DT_control=DT_control)

    def eval_at_integrator(self, stage, expr, k, i):
        DT_control = self.get_DT_control_at(k)
        DT = self.get_DT_at(k, i)
        return stage.master._method.eval_top(stage.master,
                                             stage._expr_apply(expr,
                                                               t0=self.t0,
                                                               T=self.T,
                                                               x=self.xk[k*self.M + i],
                                                               xq=self.xqk[k*self.M + i],
                                                               z=self.zk[k*self.M + i] if self.zk else nan,
                                                               u=self.U[k], p_control=self.get_p_control_at(stage, k),
                                                               p_control_plus=self.get_p_control_plus_at(stage, k),
                                                               v=self.V, p=veccat(*self.P),
                                                               v_control=self.get_v_control_at(stage, k),
                                                               v_control_plus=self.get_v_control_plus_at(stage, k),
                                                               v_states=self.get_v_states_at(stage, k),
                                                               t=self.integrator_grid[k][i],
                                                               DT=DT,
                                                               DT_control=DT_control))

    def eval_at_integrator_root(self, stage, expr, k, i, j):
        DT_control = self.get_DT_control_at(k)
        DT = self.get_DT_at(k, i)
        return stage.master._method.eval_top(stage.master,
                                             stage._expr_apply(expr,
                                                               t0=self.t0,
                                                               T=self.T,
                                                               x=self.xr[k][i][:,j],
                                                               z=self.zr[k][i][:,j] if self.zk else nan,
                                                               u=self.U[k],
                                                               p_control=self.get_p_control_at(stage, k),
                                                               p_control_plus=self.get_p_control_plus_at(stage, k),
                                                               v=self.V, p=veccat(*self.P),
                                                               v_control=self.get_v_control_at(stage, k),
                                                               v_control_plus=self.get_v_control_plus_at(stage, k),
                                                               t=self.tr[k][i][j],
                                                               DT=DT,
                                                               DT_control=DT_control))

    def set_initial(self, stage, master, initial):
        opti = master.opti if hasattr(master, 'opti') else master
        opti.cache_advanced()
        initial = HashOrderedDict(initial)
        algs = get_ranges_dict(stage.algebraics)
        initial_alg = HashDict()
        for a, v in list(initial.items()):
            if a in algs:
                initial_alg[a] = v
                del initial[a]
        for var, expr in initial.items():
            if is_equal(var, stage.T):
                var = self.T
            if is_equal(var, stage.t0):
                var = self.t0
            opti_initial = opti.initial()
            if is_numeric(expr):
                value = ca.evalf(expr)
            else:
                expr = ca.hcat([self.eval_at_control(stage, expr, k) for k in list(range(self.N))+[-1]]) # HOT line
                value = DM(opti.debug.value(expr, opti_initial))
            # Row vector if vector
            if value.is_column() and var.is_scalar(): value = value.T
            if var in self.signals:
                target = stage.sample(var,'gist')[1]
                opti.set_initial(target, ca.repmat(value,1,target.shape[1]), cache_advanced=True)
            for k in list(range(self.N))+[-1]:
                target = self.eval_at_control(stage, var, k)
                value_k = value
                if target.numel()*(self.N)==value.numel() or target.numel()*(self.N+1)==value.numel():
                    value_k = value[:,k]
                try:
                    #print(target,value_k)
                    opti.set_initial(target, value_k, cache_advanced=True)
                except Exception as e:
                    # E.g for single shooting, set_initial of a state, for k>0
                    # Error message is usually "... arbitrary expression ..." but can also be
                    # "... You cannot set an initial value for a parameter ..."
                    # if the dynamics contains a parameter
                    if "arbitrary expression" in str(e) or (not target.is_valid_input() and "initial value for a parameter" in str(e)):
                        pass
                    else:
                        # Other type of error: 
                        raise e
        for var, expr in initial_alg.items():
            opti_initial = opti.initial()
            if is_numeric(expr):
                value = ca.evalf(expr)
            else:
                expr = ca.hcat([self.eval_at_control(stage, expr, k) for k in range(self.N)]) # HOT line
                value = DM(opti.debug.value(expr, opti_initial))

            # Row vector if vector
            if value.is_column() and var.is_scalar(): value = value.T


            print("value",value)
            for k in range(self.N):
                value_k = value
                z0 = self.Z0[k]
                if z0 is None: break
                target = z0[algs[var]]
                if target.numel()*(self.N)==value.numel() or target.numel()*(self.N+1)==value.numel():
                    value_k = value[:,k]
                opti.set_value(target, value_k)

    def set_value(self, stage, master, parameter, value):
        opti = master.opti if hasattr(master, 'opti') else master
        found = False
        for i, p in enumerate(stage.parameters['']):
            if is_equal(parameter, p):
                found = True
                opti.set_value(self.P[i], value)
        for i, p in enumerate(stage.parameters['control']):
            if is_equal(parameter, p):
                found = True
                opti.set_value(hcat(self.P_control[i]), value)
        for i, p in enumerate(stage.parameters['control+']):
            if is_equal(parameter, p):
                found = True
                opti.set_value(hcat(self.P_control_plus[i]), value)
        for p in stage.parameters['bspline']:
            if is_equal(parameter, p):
                found = True
                opti.set_value(self.signals[p].coeff, value)
        assert found, "You attempted to set the value of a non-parameter."
    def add_parameter(self, stage, opti):
        for p in stage.parameters['']:
            self.P.append(opti.parameter(p.shape[0], p.shape[1]))
        for p in stage.parameters['control']:
            self.P_control.append([opti.parameter(p.shape[0], p.shape[1]) for i in range(self.N)])
        for p in stage.parameters['control+']:
            self.P_control_plus.append([opti.parameter(p.shape[0], p.shape[1]) for i in range(self.N+1)])

    def add_parameter_signals(self, stage, opti):
        # These in general depend on self.T (at least derivatives), which is not yet available in add_parameter
        for p in stage.parameters["bspline"]:
            cat = stage._catalog[p]
            # Compute the degree and size of a BSpline coefficient needed
            d = cat["order"]
            s = self.N+d
            assert p.size2()==1
            C = opti.parameter(p.size1(), s)
            BSplineSignal.register(self.signals, p, stage, BSplineSignal(C, self.xi, d, T = self.T,parametric=True))

    def set_parameter(self, stage, opti):
        for i, p in enumerate(stage.parameters['']):
            opti.set_value(self.P[i], stage._param_value(p))
        for i, p in enumerate(stage.parameters['control']):
            opti.set_value(hcat(self.P_control[i]), stage._param_value(p))
        for i, p in enumerate(stage.parameters['control+']):
            opti.set_value(hcat(self.P_control_plus[i]), stage._param_value(p))
        for p in stage.parameters['bspline']:
            opti.set_value(self.signals[p].coeff, stage._param_value(p))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Optimal control problem."""

from setuptools import setup, find_packages
import glob
import os

version = "0.2.6"

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):

        for filename in filenames:
            if filename.endswith(".pyd"): continue
            full = os.path.join(path, filename)
            if "__pycache__" in full: continue
            if "acados_matlab_octave" in full: continue
            if "examples" in full: continue
            if "/." in full: continue
            if "/test/" in full: continue
            if "/docs/" in full: continue
            if "/external/utils" in full: continue
            if "/external/external/" in full and ("hpipm" not in full and "blasfeo" not in full and "/external/external/CMakeLists.txt" not in full): continue
            paths.append(full)
    return paths

setup(
    name='rockit-meco',
    version=version,
    author="MECO-Group",
    author_email="joris.gillis@kuleuven.be",
    description="Rapid Optimal Control Kit",
    license='LICENSE',
    keywords="OCP optimal control casadi",
    url='https://gitlab.kuleuven.be/meco-software/rockit',
    packages=find_packages(exclude=['tests', 'examples']),
    include_package_data=True,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'casadi>=3.5,<4.0',
        'numpy',
        'matplotlib',
        'scipy',
        'future-fstrings'
    ],
    data_files=[('acados',package_files('rockit/external/acados/external')),
                ('acados_interface',package_files('rockit/external/acados/interface_generation'))],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering'
    ],
    download_url='https://gitlab.kuleuven.be/meco-software/rockit/-/archive/v%s/rockit-v%s.tar.gz' % (version, version)
)



# Insert pythonpath
import sys
import os

sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'external', 'interfaces', 'acados_template'))

try:
    import acados_template
except:
    raise Exception("'import acados_template' failed.\n"
                    "Check https://docs.acados.org/interfaces/index.html#python-templates for install instructions.")

from .method import AcadosMethod

method = AcadosMethod


#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#
#   Author: Jonathan Frey: jonathanpaulfrey(at)gmail.com

from .determine_input_nonlinearity_function import determine_input_nonlinearity_function
from .check_reformulation import check_reformulation
from casadi import *
from ..utils import casadi_length, idx_perm_to_ipiv, is_empty


def reformulate_with_LOS(acados_ocp, gnsf, print_info):

    ## Description:
    # This function takes an intitial transcription of the implicit ODE model
    # "model" into "gnsf" and reformulates "gnsf" with a linear output system
    # (LOS), containing as many states of the model as possible.
    # Therefore it might be that the state vector and the implicit function
    # vector have to be reordered. This reordered model is part of the output,
    # namely reordered_model.

    ## import CasADi and load models
    model = acados_ocp.model

    # symbolics
    x = gnsf["x"]
    xdot = gnsf["xdot"]
    u = gnsf["u"]
    z = gnsf["z"]

    # dimensions
    nx = gnsf["nx"]
    nz = gnsf["nz"]

    # get model matrices
    A = gnsf["A"]
    B = gnsf["B"]
    C = gnsf["C"]
    E = gnsf["E"]
    c = gnsf["c"]

    A_LO = gnsf["A_LO"]

    y = gnsf["y"]

    phi_old = gnsf["phi_expr"]

    if print_info:
        print(" ")
        print("=================================================================")
        print(" ")
        print("================    Detect Linear Output System   ===============")
        print(" ")
        print("=================================================================")
        print(" ")
    ## build initial I_x1 and I_x2_candidates
    # I_xrange( all components of x for which either xii or xdot_ii enters y):
    # I_LOS_candidates: the remaining components

    I_nsf_components = set()
    I_LOS_candidates = set()

    if gnsf["ny"] > 0:
        for ii in range(nx):
            if which_depends(y, x[ii])[0] or which_depends(y, xdot[ii])[0]:
                # i.e. xii or xiidot are part of y, and enter phi_expr
                if print_info:
                    print(f"x_{ii} is part of x1")
                I_nsf_components = set.union(I_nsf_components, set([ii]))
            else:
                # i.e. neither xii nor xiidot are part of y, i.e. enter phi_expr
                I_LOS_candidates = set.union(I_LOS_candidates, set([ii]))
                if print_info:
                    print(" ")
        for ii in range(nz):
            if which_depends(y, z[ii])[0]:
                # i.e. xii or xiidot are part of y, and enter phi_expr
                if print_info:
                    print(f"z_{ii} is part of x1")
                I_nsf_components = set.union(I_nsf_components, set([ii + nx]))
            else:
                # i.e. neither xii nor xiidot are part of y, i.e. enter phi_expr
                I_LOS_candidates = set.union(I_LOS_candidates, set([ii + nx]))
    else:
        I_LOS_candidates = set(range((nx + nz)))
    if print_info:
        print(" ")
        print(f"I_LOS_candidates {I_LOS_candidates}")
    new_nsf_components = I_nsf_components
    I_nsf_eq = set([])
    unsorted_dyn = set(range(nx + nz))
    xdot_z = vertcat(xdot, z)

    ## determine components of Linear Output System
    # determine maximal index set I_x2
    # such that the components x(I_x2) can be written as a LOS
    Eq_map = []
    while True:
        ## find equations corresponding to new_nsf_components
        for ii in new_nsf_components:
            current_var = xdot_z[ii]
            var_name = current_var.name

            # print( unsorted_dyn)
            # print("np.nonzero(E[:,ii])[0]",np.nonzero(E[:,ii])[0])
            I_eq = set.intersection(set(np.nonzero(E[:, ii])[0]), unsorted_dyn)
            if len(I_eq) == 1:
                i_eq = I_eq.pop()
                if print_info:
                    print(f"component {i_eq} is associated with state {ii}")
            elif len(I_eq) > 1:  # x_ii_dot occurs in more than 1 eq linearly
                # find the equation with least linear dependencies on
                # I_LOS_cancidates
                number_of_eq = 0
                candidate_dependencies = np.zeros(len(I_eq), 1)
                I_x2_candidates = set.intersection(I_LOS_candidates, set(range(nx)))
                for eq in I_eq:
                    depending_candidates = set.union(
                        np.nonzero(E[eq, I_LOS_candidates])[0],
                        np.nonzero(A[eq, I_x2_candidates])[0],
                    )
                    candidate_dependencies[number_of_eq] = +len(depending_candidates)
                    number_of_eq += 1
                    number_of_eq = np.argmin(candidate_dependencies)
                i_eq = I_eq[number_of_eq]
            else:  ## x_ii_dot does not occur linearly in any of the unsorted dynamics
                for j in unsorted_dyn:
                    phi_eq_j = gnsf["phi_expr"][np.nonzero(C[j, :])[0]]
                    if which_depends(phi_eq_j, xdot_z(ii))[0]:
                        I_eq = set.union(I_eq, j)
                if is_empty(I_eq):
                    I_eq = unsorted_dyn
                # find the equation with least linear dependencies on I_LOS_cancidates
                number_of_eq = 0
                candidate_dependencies = np.zeros(len(I_eq), 1)
                I_x2_candidates = set.intersection(I_LOS_candidates, set(range(nx)))
                for eq in I_eq:
                    depending_candidates = set.union(
                        np.nonzero(E[eq, I_LOS_candidates])[0],
                        np.nonzero(A[eq, I_x2_candidates])[0],
                    )
                    candidate_dependencies[number_of_eq] = +len(depending_candidates)
                    number_of_eq += 1
                    number_of_eq = np.argmin(candidate_dependencies)
                i_eq = I_eq[number_of_eq]
                ## add 1 * [xdot,z](ii) to both sides of i_eq
                if print_info:
                    print(
                        "adding 1 * ",
                        var_name,
                        " to both sides of equation ",
                        i_eq,
                        ".",
                    )
                gnsf["E"][i_eq, ii] = 1
                i_phi = np.nonzero(gnsf["C"][i_eq, :])
                if is_empty(i_phi):
                    i_phi = len(gnsf["phi_expr"]) + 1
                    gnsf["C"][i_eq, i_phi] = 1  # add column to C with 1 entry
                    gnsf["phi_expr"] = vertcat(gnsf["phi_expr"], 0)
                    gnsf["phi_expr"][i_phi] = (
                        gnsf["phi_expr"](i_phi)
                        + gnsf["E"][i_eq, ii] / gnsf["C"][i_eq, i_phi] * xdot_z[ii]
                    )
                if print_info:
                    print(
                        "detected equation ",
                        i_eq,
                        " to correspond to variable ",
                        var_name,
                    )
            I_nsf_eq = set.union(I_nsf_eq, {i_eq})
            # remove i_eq from unsorted_dyn
            unsorted_dyn.remove(i_eq)
            Eq_map.append([ii, i_eq])

        ## add components to I_x1
        for eq in I_nsf_eq:
            I_linear_dependence = set.union(
                set(np.nonzero(A[eq, :])[0]), set(np.nonzero(E[eq, :])[0])
            )
            I_nsf_components = set.union(I_linear_dependence, I_nsf_components)
            # I_nsf_components = I_nsf_components[:]

        new_nsf_components = set.intersection(I_LOS_candidates, I_nsf_components)
        if is_empty(new_nsf_components):
            if print_info:
                print("new_nsf_components is empty")
            break
        # remove new_nsf_components from candidates
        I_LOS_candidates = set.difference(I_LOS_candidates, new_nsf_components)
    if not is_empty(Eq_map):
        # [~, new_eq_order] = sort(Eq_map(1,:))
        # I_nsf_eq = Eq_map(2, new_eq_order )
        for count, m in enumerate(Eq_map):
            m.append(count)
        sorted(Eq_map, key=lambda x: x[1])
        new_eq_order = [m[2] for m in Eq_map]
        Eq_map = [Eq_map[i] for i in new_eq_order]
        I_nsf_eq = [m[1] for m in Eq_map]

    else:
        I_nsf_eq = []

    I_LOS_components = I_LOS_candidates
    I_LOS_eq = sorted(set.difference(set(range(nx + nz)), I_nsf_eq))
    I_nsf_eq = sorted(I_nsf_eq)

    I_x1 = set.intersection(I_nsf_components, set(range(nx)))
    I_z1 = set.intersection(I_nsf_components, set(range(nx, nx + nz)))
    I_z1 = set([i - nx for i in I_z1])

    I_x2 = set.intersection(I_LOS_components, set(range(nx)))
    I_z2 = set.intersection(I_LOS_components, set(range(nx, nx + nz)))
    I_z2 = set([i - nx for i in I_z2])

    if print_info:
        print(f"I_x1 {I_x1}, I_x2 {I_x2}")

    ## permute x, xdot
    if is_empty(I_x1):
        x1 = []
        x1dot = []
    else:
        x1 = x[list(I_x1)]
        x1dot = xdot[list(I_x1)]
    if is_empty(I_x2):
        x2 = []
        x2dot = []
    else:
        x2 = x[list(I_x2)]
        x2dot = xdot[list(I_x2)]
    if is_empty(I_z1):
        z1 = []
    else:
        z1 = z(I_z1)
    if is_empty(I_z2):
        z2 = []
    else:
        z2 = z[list(I_z2)]

    I_x1 = sorted(I_x1)
    I_x2 = sorted(I_x2)
    I_z1 = sorted(I_z1)
    I_z2 = sorted(I_z2)
    gnsf["xdot"] = vertcat(x1dot, x2dot)
    gnsf["x"] = vertcat(x1, x2)
    gnsf["z"] = vertcat(z1, z2)
    gnsf["nx1"] = len(I_x1)
    gnsf["nx2"] = len(I_x2)
    gnsf["nz1"] = len(I_z1)
    gnsf["nz2"] = len(I_z2)

    # store permutations
    gnsf["idx_perm_x"] = I_x1 + I_x2
    gnsf["ipiv_x"] = idx_perm_to_ipiv(gnsf["idx_perm_x"])
    gnsf["idx_perm_z"] = I_z1 + I_z2
    gnsf["ipiv_z"] = idx_perm_to_ipiv(gnsf["idx_perm_z"])
    gnsf["idx_perm_f"] = I_nsf_eq + I_LOS_eq
    gnsf["ipiv_f"] = idx_perm_to_ipiv(gnsf["idx_perm_f"])

    f_LO = SX.sym("f_LO", 0, 0)

    ## rewrite I_LOS_eq as LOS
    if gnsf["n_out"] == 0:
        C_phi = np.zeros(gnsf["nx"] + gnsf["nz"], 1)
    else:
        C_phi = C @ phi_old
    if gnsf["nx1"] == 0:
        Ax1 = np.zeros(gnsf["nx"] + gnsf["nz"], 1)
    else:
        Ax1 = A[:, sorted(I_x1)] @ x1
    if gnsf["nx1"] + gnsf["nz1"] == 0:
        lhs_nsf = np.zeros(gnsf["nx"] + gnsf["nz"], 1)
    else:
        lhs_nsf = E[:, sorted(I_nsf_components)] @ vertcat(x1, z1)
    n_LO = len(I_LOS_eq)
    B_LO = np.zeros((n_LO, gnsf["nu"]))
    A_LO = np.zeros((gnsf["nx2"] + gnsf["nz2"], gnsf["nx2"]))
    E_LO = np.zeros((n_LO, n_LO))
    c_LO = np.zeros((n_LO, 1))

    I_LOS_eq = list(I_LOS_eq)
    for eq in I_LOS_eq:
        i_LO = I_LOS_eq.index(eq)
        f_LO = vertcat(f_LO, Ax1[eq] + C_phi[eq] - lhs_nsf[eq])
        print(f"eq {eq} I_LOS_components {I_LOS_components}, i_LO {i_LO}, f_LO {f_LO}")
        E_LO[i_LO, :] = E[eq, sorted(I_LOS_components)]
        A_LO[i_LO, :] = A[eq, I_x2]
        c_LO[i_LO, :] = c[eq]
        B_LO[i_LO, :] = B[eq, :]
    if casadi_length(f_LO) == 0:
        f_LO = SX.zeros((gnsf["nx2"] + gnsf["nz2"], 1))
    f_LO = simplify(f_LO)
    gnsf["A_LO"] = A_LO
    gnsf["E_LO"] = E_LO
    gnsf["B_LO"] = B_LO
    gnsf["c_LO"] = c_LO
    gnsf["f_lo_expr"] = f_LO

    ## remove I_LOS_eq from NSF type system
    gnsf["A"] = gnsf["A"][np.ix_(sorted(I_nsf_eq), sorted(I_x1))]
    gnsf["B"] = gnsf["B"][sorted(I_nsf_eq), :]
    gnsf["C"] = gnsf["C"][sorted(I_nsf_eq), :]
    gnsf["E"] = gnsf["E"][np.ix_(sorted(I_nsf_eq), sorted(I_nsf_components))]
    gnsf["c"] = gnsf["c"][sorted(I_nsf_eq), :]

    ## reduce phi, C
    I_nonzero = []
    for ii in range(gnsf["C"].shape[1]):  # n_colums of C:
        print(f"ii {ii}")
        if not all(gnsf["C"][:, ii] == 0):  # if column ~= 0
            I_nonzero.append(ii)
    gnsf["C"] = gnsf["C"][:, I_nonzero]
    gnsf["phi_expr"] = gnsf["phi_expr"][I_nonzero]

    gnsf = determine_input_nonlinearity_function(gnsf)

    check_reformulation(model, gnsf, print_info)

    gnsf["nontrivial_f_LO"] = 0
    if not is_empty(gnsf["f_lo_expr"]):
        for ii in range(casadi_length(gnsf["f_lo_expr"])):
            fii = gnsf["f_lo_expr"][ii]
            if not fii.is_zero():
                gnsf["nontrivial_f_LO"] = 1
            if not gnsf["nontrivial_f_LO"] and print_info:
                print("f_LO is fully trivial (== 0)")
    check_reformulation(model, gnsf, print_info)

    if print_info:
        print("")
        print(
            "---------------------------------------------------------------------------------"
        )
        print(
            "------------- Success: Linear Output System (LOS) detected ----------------------"
        )
        print(
            "---------------------------------------------------------------------------------"
        )
        print("")
        print(
            "==>>  moved  ",
            gnsf["nx2"],
            "differential states and ",
            gnsf["nz2"],
            " algebraic variables to the Linear Output System",
        )
        print(
            "==>>  recuced output dimension of phi from  ",
            casadi_length(phi_old),
            " to ",
            casadi_length(gnsf["phi_expr"]),
        )
        print(" ")
        print("Matrices defining the LOS read as")
        print(" ")
        print("E_LO =")
        print(gnsf["E_LO"])
        print("A_LO =")
        print(gnsf["A_LO"])
        print("B_LO =")
        print(gnsf["B_LO"])
        print("c_LO =")
        print(gnsf["c_LO"])

    return gnsf



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from acados_template import AcadosOcp, AcadosOcpSolver, AcadosSimSolver, AcadosModel
from acados_template.utils import J_to_idx

from ...freetime import FreeTime
from ..method import ExternalMethod, legit_J, check_Js
from ...solution import OcpSolution

import numpy as np
from casadi import SX, Sparsity, MX, vcat, veccat, symvar, substitute, sparsify, DM, Opti, is_linear, vertcat, depends_on, jacobian, linear_coeff, quadratic_coeff, mtimes, pinv, evalf, Function, vvcat, inf, sum1, sum2, diag
import casadi
from ...casadi_helpers import DM2numpy, reshape_number
from collections import OrderedDict
import os
import shutil
import subprocess
import casadi as ca
import tempfile

INF = 1e5

def linear_coeffs(expr, *args):
    """ Multi-argument extesion to CasADi linear_coeff"""
    J,c = linear_coeff(expr, vcat(args))
    cs = np.cumsum([0]+[e.numel() for e in args])
    return tuple([J[:,cs[i]:cs[i+1]] for i in range(len(args))])+(c,)


def legit_Js(J):
    """
    Checks if J, a pre-multiplier for slacks, is of legitimate structure
    Empty rows are allowed
    """
    try:
        check_Js(J)
    except:
        return False
    return True


"""
    def __init__(self,**kwargs):
        ExternalMethod.__init__(self, **kwargs)
        self.build_dir = tempfile.TemporaryDirectory(prefix="acados_rockit_")
        self.build_dir_abs = self.build_dir.name
        
    def __del__(self):
        self.build_dir.cleanup()
"""

def recursive_overwrite(src, dest, ignore=None):
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        if ignore is not None:
            ignored = ignore(src, files)
        else:
            ignored = set()
        for f in files:
            if f not in ignored:
                recursive_overwrite(os.path.join(src, f), 
                                    os.path.join(dest, f), 
                                    ignore)
    else:
        shutil.copyfile(src, dest, follow_symlinks=False)

class AcadosMethod(ExternalMethod):
    def __init__(self,feasibility_problem=False,acados_options=None,model_name="rockit_model",**kwargs):
        ExternalMethod.__init__(self, **kwargs)
        #self.build_dir_abs = "./build_acados_rockit"
        self.build_dir_abs = "./foobar"

        self.model_name = model_name

        self.feasibility_problem = feasibility_problem
        self.acados_options = {} if acados_options is None else acados_options

    def fill_placeholders_integral(self, phase, stage, expr, *args):
        if phase==1:
            I = stage.state()
            stage.set_der(I, expr)
            stage.subject_to(stage.at_t0(I)==0)
            return stage.at_tf(I)

    def transcribe_phase1(self, stage, **kwargs):
        self.stage = stage
        self.ocp = ocp = AcadosOcp()
        self.opti = Opti()


        self.X_gist = [ca.vvcat([MX.sym("Xg", s.size1(), s.size2()) for s in stage.states]) for k in range(self.N+1)]
        self.U_gist = [ca.vvcat([MX.sym("Ug", s.size1(), s.size2()) for s in stage.controls]) for k in range(self.N)]
        self.P_stage_gist = [ca.vvcat([MX.sym("stage_param%d" % k, s.size1(), s.size2()) for s in  self.p_local]) for k in range(self.N+1)]
        self.P_global_gist = ca.vvcat([MX.sym("Vg", s.size1(), s.size2()) for s in stage.parameters['']])
        self.T_gist = MX.sym("Tg")


        self.gist_parts = []
        self.gist_parts.append((self.X_gist, stage.x, "local"))
        self.gist_parts.append((self.U_gist, stage.u, "local"))
        self.gist_parts.append((self.P_global_gist, self.p_global_cat, "global"))
        self.gist_parts.append((self.P_stage_gist, self.p_local_cat, "local"))
        self.gist_parts.append((self.T_gist, stage.T, "global"))

        import gc
        gc.collect()

        model = AcadosModel()

        f = stage._ode()

        if self.expand:
            model_x = SX.sym("x", stage.x.sparsity())
            model_u = SX.sym("u", stage.u.sparsity())
            model_p = SX.sym("p", stage.p.sparsity())
            model_xdot = SX.sym("xdot", stage.x.sparsity())

            tT = Function('temp',[stage.x,stage.u,stage.p],[self.t, self.T])
            model_t, model_T = tT(model_x, model_u, model_p)
        else:
            model_x = stage.x
            model_u = stage.u
            model_p = stage.p
            model_xdot = MX.sym("xdot", stage.nx)
            model_t = self.t
            model_T = self.T

        res = f(x=model_x, u=model_u, p=vertcat(model_p, DM.nan(stage.v.shape)), t=model_t)
        if isinstance(stage._T,FreeTime):
            f_expl = model_T*res["ode"]
        else:
            f_expl = res["ode"]

        model.f_impl_expr = model_xdot-f_expl
        model.f_expl_expr = f_expl
        model.x = model_x
        model.xdot = model_xdot
        model.u = model_u
        model.p = model_p
        self.P0 = DM.zeros(stage.np)

        slack = MX(0, 1) if len(stage.variables['control'])==0 else vvcat(stage.variables['control'])

        slack_bounds = stage.variables['']
        slack_0_index = set()
        slack_e_index = set()

        for c, meta, _ in stage._constraints["point"]:
            sv = symvar(c)
            case_0 = np.any(["_t0" in e.name() for e in sv])
            case_e = np.any(["_tf" in e.name() for e in sv])
            assert not (case_0 and case_e), "Point constraints must be either at t0 or tf"

            for i,s_b in enumerate(slack_bounds):
                if np.any([ca.is_equal(s_b,e) for e in sv]):
                    if case_0:
                        assert i not in slack_0_index, "Slack can only be used once in a point constraint"
                        slack_0_index.add(i)
                    elif case_e:
                        assert i not in slack_e_index, "Slack can only be used once in a point constraint"
                        slack_e_index.add(i)
        slack_0 = [slack_bounds[i] for i in sorted(slack_0_index)]
        slack_e = [slack_bounds[i] for i in sorted(slack_e_index)]

        slack_0 = MX(0, 1) if len(slack_0)==0 else vvcat(slack_0)
        slack_e = MX(0, 1) if len(slack_e)==0 else vvcat(slack_e)

        self.slack = slack
        self.slack_0 = slack_0
        self.slack_e = slack_e

        self.X = self.opti.variable(*stage.x.shape)
        self.U = self.opti.variable(*stage.u.shape)
        self.P = self.opti.parameter(*stage.p.shape)
        self.S = self.opti.variable(*slack.shape)
        self.Se = self.opti.variable(*slack_e.shape)
        self.S0 = self.opti.variable(*slack_0.shape)
        self.t = self.opti.parameter()

        self.raw = [stage.x,stage.u,stage.p,slack,slack_0,slack_e,stage.t]
        self.optivar = [self.X, self.U, self.P, self.S, self.S0, self.Se, self.t]

        #self.time_grid = self.grid(stage._t0, stage._T, self.N)
        self.normalized_time_grid = self.grid(0.0, 1.0, self.N)
        self.time_grid = self.normalized_time_grid
        if not isinstance(stage._T, FreeTime): self.time_grid*= stage._T
        if not isinstance(stage._t0, FreeTime): self.time_grid+= stage._t0
        self.control_grid = MX(stage.t0 + self.normalized_time_grid*stage.T).T

        inits = []
        inits.append((stage.T, stage._T.T_init if isinstance(stage._T, FreeTime) else stage._T))
        inits.append((stage.t0, stage._t0.T_init if isinstance(stage._t0, FreeTime) else stage._t0))

        self.inits = inits

        self.control_grid_init = evalf(substitute([self.control_grid], [a for a,b in inits],[b for a,b in inits])[0])

        #self.control_grid = self.normalized_time_grid

        model.name = self.model_name

        ocp.model = model

        # set dimensions
        ocp.dims.N = self.N

        lagrange = MX(0)
        mayer = MX(0)
        mayer0 = MX(0)
        
        def split_sum(e):
            if e.op()==ca.OP_ADD:
                for p in split_sum(e.dep(0)):
                    yield p
                for p in split_sum(e.dep(1)):
                    yield p
            else:
                yield e

        for obj in split_sum(MX(stage._objective)):
            if "sum_control" in str(obj):
                lagrange += obj
            elif ca.depends_on(obj,slack_0) or "_at_t0" in str(obj):
                mayer0 += obj
            else:
                mayer += obj

        self.lagrange = lagrange
        self.mayer = mayer
        self.mayer0 = mayer0

        self.initial_conditions = []
        self.final_conditions = []

    def transcribe_phase2(self, stage, **kwargs):
        opti_advanced = self.opti.advanced
        placeholders = kwargs["placeholders"]

        # Total Lagrange integrand
        lagrange = placeholders(self.lagrange,preference=['expose'])*self.N
        # Total Mayer term
        mayer = placeholders(self.mayer,preference=['expose'])
        mayer0 = placeholders(self.mayer0,preference=['expose'])

        assert not depends_on(mayer, stage.u), "Mayer term of objective may not depend on controls"

        ocp = self.ocp

        # At the moment, we don't detect any structure in cost...
        #ocp.cost.cost_type_0 = 'EXTERNAL'
        ocp.cost.cost_type = 'EXTERNAL'
        ocp.cost.cost_type_e = 'EXTERNAL'

        # .. except for slacks
        assert not depends_on(lagrange, vertcat(self.slack_0,self.slack_e)), "Lagrange term may not depend on a non-signal slack"
        assert not depends_on(mayer, self.slack), "Mayer term may not depend on a signal slack"
        assert not depends_on(mayer0, self.slack), "Mayer0 term may not depend on a signal slack"
        assert not depends_on(mayer, self.slack_0), "Mayer term may not depend on a slack_0"
        assert not depends_on(mayer0, self.slack_e), "Mayer0 term may not depend on a slack_e"

        Qs, bs, lagrange = quadratic_coeff(lagrange, self.slack)
        Qs += DM.zeros(Sparsity.diag(Qs.shape[0]))
        assert Qs.sparsity().is_diag(), "Slacks cannot be mixed in Lagrange objective"
        Qs_e, bs_e, mayer = quadratic_coeff(mayer, vertcat(self.slack,self.slack_e))
        Qs_e += DM.zeros(Sparsity.diag(Qs_e.shape[0]))
        assert Qs_e.sparsity().is_diag(), "Slacks cannot be mixed in Mayer objective"

        Qs_0, bs_0, mayer0 = quadratic_coeff(mayer0, vertcat(self.slack,self.slack_0))
        Qs_0 += DM.zeros(Sparsity.diag(Qs_0.shape[0]))
        assert Qs_0.sparsity().is_diag(), "Slacks cannot be mixed in Mayer0 objective"

        assert not depends_on(veccat(Qs, bs, Qs_0, bs_0, Qs_e, bs_e), vertcat(stage.x, stage.u, self.slack, self.slack_0, self.slack_e)), \
            "Slack part of objective must be quadratic in slacks and depend only on parameters"

        #ocp.model.cost_expr_ext_cost_0 = mayer0
        ocp.model.cost_expr_ext_cost = lagrange
        ocp.model.cost_expr_ext_cost_e = mayer

        #import ipdb
        #ipdb.set_trace()
        #raise Exception("")

        # For reference: https://github.com/acados/acados/blob/master/docs/problem_formulation/problem_formulation_ocp_mex.pdf

        Jbx = []; lbx = []; ubx = []; Jsbx = []
        Jbu = []; lbu = []; ubu = []; Jsbu = []
        C = []; D = []; lg = []; ug = []; Jsg = []
        h = []; lh = []; uh = []; Jsh = []

        h_e = []; lh_e = []; uh_e = []; Jsh_e = []
        h_0 = []; lh_0 = []; uh_0 = []; Jsh_0 = []
        Jbx_e = []; lbx_e = []; ubx_e = []; Jsbx_e = []
        C_e = []; lg_e = []; ug_e = []; Jsg_e = []

        lbx_0 = [];ubx_0 = []
        Jbx_0 = [];Jbxe_0 = []



        def mark(slack, Js):
            assert np.all(np.array(slack[Js.sparsity().get_col()])==0)
            slack[Js.sparsity().get_col()] = 1
    
        def export_expr(m):
            if isinstance(m,list):
                if len(m)==0:
                    return MX(0, 1)
                else:
                    return vcat(m)
            return m
        
        def export_num(m):
            res=np.array(evalf(export_expr(m)))
            if np.any(res==-inf) or np.any(res==inf):
                print("WARNING: Double-sided constraints are much preferred. Replaced inf with %f." % INF)
            res[res==-inf] = -INF
            res[res==inf] = INF
            return res

        def export_num_vec(m):
            return np.array(evalf(export_expr(m))).reshape(-1)

        def export(m):
            return (export_expr(m),False)

        def export_vec(m):
            return (export_expr(m),True)


        # Flags to verify that each slack has a >=0 constraint
        slack_has_pos_const = DM.zeros(*self.slack.shape)

        # Flags to check sign
        slack_lower = DM.zeros(*self.slack.shape)
        slack_upper = DM.zeros(*self.slack.shape)

        # Process path constraints
        for c, meta, args in stage._constraints["control"]+stage._constraints["integrator"]:
            try:
                c = substitute([placeholders(c,preference=['expose'])],self.raw,self.optivar)[0]
                mc = opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
                lb,canon,ub = substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw)

                # Check for infinities
                try:
                    lb_inf = np.all(np.array(evalf(lb)==-inf))
                except:
                    lb_inf = False
                try:
                    ub_inf = np.all(np.array(evalf(ub)==inf))
                except:
                    ub_inf = False

                assert not depends_on(canon, self.slack_e), "Path constraints may only have signal slacks"
                
                # Slack positivity constraint
                if depends_on(canon, self.slack) and not depends_on(canon, vertcat(stage.x, stage.u, stage.p)):
                    J,c = linear_coeff(canon, self.slack)
                    is_perm = legit_Js(J)
                    lb_zero = np.all(np.array(evalf(lb-c)==0))
                    assert is_perm and lb_zero and ub_inf, "Only constraints allowed on slacks are '>=0'"
                    slack_has_pos_const[J.sparsity().get_col()] = 1
                    continue
                
                assert is_linear(canon, self.slack), "slacks can only enter linearly in constraints"

                if is_linear(canon, vertcat(stage.x,stage.u)):
                    # Linear is states and controls
                    if not depends_on(canon, stage.x):
                        # lbu <= Jbu u <= ubu
                        J, Js, c = linear_coeffs(canon, stage.u, self.slack)
                        if legit_J(J):
                            Jbu.append(J)
                            lbu.append(lb-c)
                            ubu.append(ub-c)
                            if Js.nnz()>0:
                                assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY]
                            
                            if not ub_inf: Js *= -1
                            check_Js(Js)
                            Jsbu.append(Js)
                            mark(slack_lower if ub_inf else slack_upper, Js)
                            continue
                        # Fallthrough
                    if not depends_on(canon, stage.u):
                        # lbx <= Jbx x <= ubx
                        J, Js, c = linear_coeffs(canon, stage.x, self.slack)
                        if legit_J(J):
                            Jbx.append(J)
                            lbx.append(lb-c)
                            ubx.append(ub-c)
                            if args["include_last"]:
                                assert not depends_on(canon, self.slack), " lbx <= Jbx x <= ubx does not support slacks for qualifier include_last=True."
                                Jbx_e.append(J)
                                lbx_e.append(lb-c)
                                ubx_e.append(ub-c)
                            if args["include_first"]:
                                assert not depends_on(canon, self.slack), "lbx <= Jbx x <= ubx does not support slacks for qualifier include_first=True."
                                Jbx_0.append(J)
                                lbx_0.append(lb-c)
                                ubx_0.append(ub-c)
                                Jbxe_0.append( (mc.type == casadi.OPTI_EQUALITY) * J)
                            if Js.nnz()>0:
                                assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY]      
                            if not ub_inf: Js *= -1
                            check_Js(Js)
                            Jsbx.append(Js)
                            mark(slack_lower if ub_inf else slack_upper, Js)
                            continue
                        # Fallthrough
                    # lg <= Cx + Du <= ug
                    J1, J2, Js, c = linear_coeffs(canon, stage.x,stage.u, self.slack)
                    C.append(J1)
                    D.append(J2)
                    lg.append(lb-c)
                    ug.append(ub-c)
                    if args["include_last"]:
                        assert not depends_on(canon, self.slack), "lg <= Cx + Du <= ug does not support slacks for qualifier include_last=True.\n" + str(meta)
                        if J2.nnz()>0:
                            raise Exception("lg <= Cx + Du <= ug only supported for qualifier include_last=False.\n"  + str(meta))
                        else:
                            C_e.append(J1)
                            lg_e.append(lb-c)
                            ug_e.append(ub-c)
                    if not args["include_first"]:
                        raise Exception("lg <= Cx + Du <= ug only supported for qualifier include_first=True.\n" + str(meta))
                    if Js.nnz()>0:
                        assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY]
                    if not ub_inf: Js *= -1
                    check_Js(Js)
                    Jsg.append(Js)
                    mark(slack_lower if ub_inf else slack_upper, Js)

                else:
                    # Nonlinear constraints
                    Js, c = linear_coeffs(canon, self.slack)
                    # lh <= h(x,u) <= uh
                    h.append(c)
                    lh.append(lb)
                    uh.append(ub)
                    if args["include_last"]:
                        if depends_on(canon, stage.u):
                            if args["include_last"]!="auto":
                                raise Exception("lh <= h(x,u) <= uh only supported for qualifier include_last=False.\n" + str(meta))
                        else:
                            assert not depends_on(canon, self.slack), "lh <= h(x,u) <= uh does not support slacks for qualifier include_last=True."
                            h_e.append(c)
                            lh_e.append(lb)
                            uh_e.append(ub)
                    if args["include_first"]:
                        assert not depends_on(canon, self.slack), "lh <= h(x,u) <= uh does not support slacks for qualifier include_first=True."
                        h_0.append(c)
                        lh_0.append(lb)
                        uh_0.append(ub)
                    if args["include_first"]:
                        raise Exception("lh <= h(x,u) <= uh only supported for qualifier include_first=False.\n" + str(meta))                            
                    if Js.nnz()>0:
                        assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY]
                    if not ub_inf: Js *= -1
                    check_Js(Js)
                    Jsh.append(Js)
                    mark(slack_lower if ub_inf else slack_upper, Js)
            except Exception as e:
                print(meta)
                raise e
        
        # Lump together Js* across individual path constraint categories
        Jsbu = MX(0, 1) if len(Jsbu)==0 else vcat(Jsbu)
        Jsbx = MX(0, 1) if len(Jsbx)==0 else vcat(Jsbx)
        Jsg = MX(0, 1) if len(Jsg)==0 else vcat(Jsg)
        Jsh = MX(0, 1) if len(Jsh)==0 else vcat(Jsh)

        # Indices needed to pull lower and upper parts of Qs, bs apart
        li = sparsify(slack_lower).sparsity().row()
        ui = sparsify(slack_upper).sparsity().row()

        ns = self.slack.nnz()

        Zl = MX(ns, ns)
        Zu = MX(ns, ns)
        zl = MX(ns, 1)
        zu = MX(ns, 1)
        Zl[li,li] = Qs[li,li]
        Zu[ui,ui] = Qs[ui,ui]
        zl[li,0] = bs[li,0]
        zu[ui,0] = bs[ui,0]

        # Columns give you the indices of used slacks
        bui = Jsbu.sparsity().get_col()
        bxi = Jsbx.sparsity().get_col()
        gi = Jsg.sparsity().get_col()
        hi = Jsh.sparsity().get_col()
        ni = bui+bxi+gi+hi

        ni0 = bui+gi
        Zl0 = Zl[ni0,ni0]
        Zu0 = Zu[ni0,ni0]
        zl0 = zl[ni0,0]
        zu0 = zu[ni0,0]

        # Re-order slacks according to (bu,bx,g,h)
        Zl = Zl[ni,ni]
        Zu = Zu[ni,ni]
        zl = zl[ni]
        zu = zu[ni]

        assert np.all(np.array(slack_has_pos_const)), "Only variables allowed are slacks (and they need '>=0' constraints)"

        # After re-ordering slacks, Js* become unit matrices interwoven with zero rows
        # But let's just work with idxs* directly
        self.ocp.constraints.idxsbu = np.array(Jsbu.sparsity().row())
        self.ocp.constraints.idxsbx = np.array(Jsbx.sparsity().row())
        self.ocp.constraints.idxsg = np.array(Jsg.sparsity().row())
        self.ocp.constraints.idxsh = np.array(Jsh.sparsity().row())

        # These should become parametric
        self.ocp.cost.Zl = export_num_vec(diag(Zl))
        self.ocp.cost.Zu = export_num_vec(diag(Zu))
        self.ocp.cost.zu = export_num_vec(zu)
        self.ocp.cost.zl = export_num_vec(zl)

        # Flags to verify that each slack has a >=0 constraint
        slack_e_has_pos_const = DM.zeros(*self.slack_e.shape)
        slack_0_has_pos_const = DM.zeros(*self.slack_0.shape)

        # Flags to check sign
        slack_e_lower = DM.zeros(*self.slack_e.shape)
        slack_e_upper = DM.zeros(*self.slack_e.shape)
        slack_0_lower = DM.zeros(*self.slack_0.shape)
        slack_0_upper = DM.zeros(*self.slack_0.shape)

        # Process point constraints
        # Probably should de-duplicate stuff wrt path constraints code
        for c, meta, _ in stage._constraints["point"]:
            try:
                # Make sure you resolve u to r_at_t0/r_at_tf
                c = placeholders(c,max_phase=1)
                has_t0 = 'r_at_t0' in [a.name() for a in symvar(c)]
                has_tf = 'r_at_tf' in [a.name() for a in symvar(c)]

                cb = c
                c = substitute([placeholders(c,preference='expose')],self.raw,self.optivar)[0]
                mc = opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
                lb,canon,ub = substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw)

                # Check for infinities
                try:
                    lb_inf = np.all(np.array(evalf(lb)==-inf))
                except:
                    lb_inf = False
                try:
                    ub_inf = np.all(np.array(evalf(ub)==inf))
                except:
                    ub_inf = False

                if has_t0 and has_tf:
                    raise Exception("Constraints mixing value at t0 and tf not allowed.")
                
                assert not depends_on(canon, self.slack), "Path constraints may only have non-signal slacks"

                if has_t0:
                    assert not depends_on(canon, self.slack_e), "Initial constraints may not have final slacks."
                if has_tf:
                    assert not depends_on(canon, self.slack_0), "Initial constraints may not have initial slacks."

                skip = False
                for slack, slack_has_pos_const in zip([self.slack_0,self.slack_e],[slack_0_has_pos_const,slack_e_has_pos_const]):
                    # Slack positivity constraint
                    if depends_on(canon, slack) and not depends_on(canon, vertcat(stage.x, stage.u, stage.p)):
                        J,c = linear_coeff(canon, slack)
                        is_perm = legit_Js(J)
                        lb_zero = np.all(np.array(evalf(lb-c)==0))
                        assert is_perm and lb_zero and ub_inf, "Only constraints allowed on slacks are '>=0'"
                        slack_has_pos_const[J.sparsity().get_col()] = 1
                        skip = True
                # Do not process these further
                if skip: continue

                if has_t0:
                    slack = self.slack_0
                if has_tf:
                    slack = self.slack_e

                
                assert is_linear(canon, slack), "slacks can only enter linearly in constraints"

                if has_t0:
                    if mc.type==casadi.OPTI_EQUALITY:
                        # t0
                        check = is_linear(canon, stage.x)
                        check = check and not depends_on(canon, stage.u)
                        assert check, "at t=t0, only equality constraints on x are allowed. Got '%s'" % str(c)

                        J,c = linear_coeff(canon, stage.x)
                        Jbx_0.append(J)
                        lbx_0.append(lb-c)
                        ubx_0.append(ub-c)

                        Jbxe_0.append( J)
                    else:
                        Js, c = linear_coeff(canon, self.slack_0)
                        # lh <= h(x,u) <= uh
                        h_0.append(c)
                        lh_0.append(lb)
                        uh_0.append(ub)
                        if not ub_inf: Js *= -1
                        check_Js(Js)
                        Jsh_e.append(Js)
                        mark(slack_0_lower if ub_inf else slack_0_upper, Js)
                else:
                    # tf
                    assert not depends_on(canon,stage.u), "Terminal constraints cannot depend on u"
                    if is_linear(canon, stage.x):
                        # lbx <= Jbx x <= ubx
                        J,Js,c = linear_coeffs(canon, stage.x, self.slack_e)
                        if legit_J(J):
                            Jbx_e.append(J)
                            lbx_e.append(lb-c)
                            ubx_e.append(ub-c)

                            if not ub_inf: Js *= -1
                            check_Js(Js)
                            Jsbx_e.append(Js)
                            mark(slack_e_lower if ub_inf else slack_e_upper, Js)

                            continue
                        # lg <= Cx <= ug
                        J, Js, c = linear_coeffs(canon, stage.x, self.slack_e)
                        C_e.append(J)
                        lg_e.append(lb-c)
                        ug_e.append(ub-c)
                        if not ub_inf: Js *= -1
                        check_Js(Js)
                        Jsg_e.append(Js)
                        mark(slack_e_lower if ub_inf else slack_e_upper, Js)
                    else:
                        Js, c = linear_coeff(canon, self.slack_e)
                        # lh <= h(x,u) <= uh
                        h_e.append(c)
                        lh_e.append(lb)
                        uh_e.append(ub)
                        if not ub_inf: Js *= -1
                        check_Js(Js)
                        Jsh_e.append(Js)
                        mark(slack_e_lower if ub_inf else slack_e_upper, Js)
            except Exception as e:
                print(meta)
                raise e

        # Lump together Js* across individual path constraint categories
        Jsbx_e = MX(0, 1) if len(Jsbx_e)==0 else vcat(Jsbx_e)
        Jsg_e = MX(0, 1) if len(Jsg_e)==0 else vcat(Jsg_e)
        Jsh_e = MX(0, 1) if len(Jsh_e)==0 else vcat(Jsh_e)

        # Indices needed to pull lower and upper parts of Qs, bs apart
        li = sparsify(slack_e_lower).sparsity().row()
        ui = sparsify(slack_e_upper).sparsity().row()

        ns = self.slack_e.nnz()

        Zl = MX(ns, ns)
        Zu = MX(ns, ns)
        zl = MX(ns, 1)
        zu = MX(ns, 1)
        Zl[li,li] = Qs_e[li,li]
        Zu[ui,ui] = Qs_e[ui,ui]
        zl[li,0] = bs_e[li,0]
        zu[ui,0] = bs_e[ui,0]

        bxi = Jsbx_e.sparsity().get_col()
        gi = Jsg_e.sparsity().get_col()
        hi = Jsh_e.sparsity().get_col()
        ni = bxi+gi+hi

        # Re-order slacks according to (bu,bx,g,h)
        Zl = Zl[ni,ni]
        Zu = Zu[ni,ni]
        zl = zl[ni]
        zu = zu[ni]

        assert np.all(np.array(slack_e_has_pos_const)), "Only variables allowed are slacks (and they need '>=0' constraints)"

        # After re-ordering slacks, Js* become unit matrices interwoven with zero rows
        # But let's just work with idxs* directly
        self.ocp.constraints.idxsbx_e = np.array(Jsbx_e.sparsity().row())
        self.ocp.constraints.idxsg_e = np.array(Jsg_e.sparsity().row())
        self.ocp.constraints.idxsh_e = np.array(Jsh_e.sparsity().row())

        # These should become parametric
        self.ocp.cost.Zl_e = export_num_vec(diag(Zl))
        self.ocp.cost.Zu_e = export_num_vec(diag(Zu))
        self.ocp.cost.zu_e = export_num_vec(zu)
        self.ocp.cost.zl_e = export_num_vec(zl)


        # Lump together Js* across individual path constraint categories
        Jsh_0 = MX(0, 1) if len(Jsh_0)==0 else vcat(Jsh_0)

        # Indices needed to pull lower and upper parts of Qs, bs apart
        li = sparsify(slack_0_lower).sparsity().row()
        ui = sparsify(slack_0_upper).sparsity().row()

        ns = self.slack_0.nnz()

        Zl = MX(ns, ns)
        Zu = MX(ns, ns)
        zl = MX(ns, 1)
        zu = MX(ns, 1)
        Zl[li,li] = Qs_0[li,li]
        Zu[ui,ui] = Qs_0[ui,ui]
        zl[li,0] = bs_0[li,0]
        zu[ui,0] = bs_0[ui,0]

        hi = Jsh_0.sparsity().get_col()
        ni = hi

        # Re-order slacks according to (h)
        Zl = Zl[ni,ni]
        Zu = Zu[ni,ni]
        zl = zl[ni]
        zu = zu[ni]

        Zl = ca.diagcat(Zl0,Zl)
        Zu = ca.diagcat(Zu0,Zu)
        zl = ca.vertcat(zl0,zl)
        zu = ca.vertcat(zu0,zu)

        assert np.all(np.array(slack_0_has_pos_const)), "Only variables allowed are slacks (and they need '>=0' constraints)"

        # After re-ordering slacks, Js* become unit matrices interwoven with zero rows
        # But let's just work with idxs* directly

        self.ocp.constraints.idxsh_0 = np.array(Jsh_0.sparsity().row())

        # These should become parametric
        self.ocp.cost.Zl_0 = export_num_vec(diag(Zl))
        self.ocp.cost.Zu_0 = export_num_vec(diag(Zu))
        self.ocp.cost.zu_0 = export_num_vec(zu)
        self.ocp.cost.zl_0 = export_num_vec(zl)


        ocp.constraints.constr_type = 'BGH'

        # No export_num here, let's do things parametrically
        self.m = m = OrderedDict()

        self.ocp.constraints.Jbx = export_num(Jbx)
        m["lbx"] = export_vec(lbx)
        m["ubx"] = export_vec(ubx)

        self.ocp.constraints.Jbu = export_num(Jbu)
        m["lbu"] = export_vec(lbu)
        m["ubu"] = export_vec(ubu)

        m["C"] = export(C)
        m["D"] = export(D)
        m["lg"] = export_vec(lg)
        m["ug"] = export_vec(ug)

        ocp.model.con_h_expr = export_expr(h)
        m["lh"] = export_vec(lh)
        m["uh"] = export_vec(uh)

        if h_0:
            ocp.model.con_h_expr_0 = export_expr(h_0)
            m["lh_0"] = export_vec(lh_0)
            m["uh_0"] = export_vec(uh_0)
        else:
            m["lh_0"] = export_vec(MX.zeros(0,1))
            m["uh_0"] =export_vec( MX.zeros(0,1))

        if h_e:
            ocp.model.con_h_expr_e = export_expr(h_e)
            m["lh_e"] = export_vec(lh_e)
            m["uh_e"] = export_vec(uh_e)
        else:
            m["lh_e"] = export_vec(MX.zeros(0,1))
            m["uh_e"] =export_vec( MX.zeros(0,1))


        self.ocp.constraints.Jbx_e = export_num(Jbx_e)
        m["lbx_e"] = export_vec(lbx_e)
        m["ubx_e"] = export_vec(ubx_e)

        m["C_e"] = export(C_e)
        m["lg_e"] = export_vec(lg_e)
        m["ug_e"] = export_vec(ug_e)

        self.ocp.constraints.Jbx_0 = export_num(Jbx_0)

        def None2Empty(a):
            if a is None:
                return MX(0, 1)
            else:
                return a

        if self.expand:
            temp = Function('temp', [stage.x, stage.u, stage.p], [None2Empty(ocp.model.cost_expr_ext_cost), None2Empty(ocp.model.cost_expr_ext_cost_e),  None2Empty(ocp.model.con_h_expr_0), None2Empty(ocp.model.con_h_expr_e), None2Empty(ocp.model.con_h_expr)])
            [ocp.model.cost_expr_ext_cost, ocp.model.cost_expr_ext_cost_e, ocp.model.con_h_expr_0, ocp.model.con_h_expr_e, ocp.model.con_h_expr] = temp(self.ocp.model.x, self.ocp.model.u, self.ocp.model.p)

        # Issue: if you indicate some states for equality, they are eliminated,
        # and the presence of other bounds on that state are problematic
        #
        # Filter idxbxe_0
        idxbxe_0 = np.sum(export_num(Jbxe_0),1).nonzero()[0]
        idx = list(J_to_idx(export_num(Jbx_0)))
        idxbxe_0_filtered = []
        for e in idxbxe_0:
            if idx.count(e)==1:
                idxbxe_0_filtered.append(e)
        # Still not safe
        idxbxe_0_filtered = []
        self.ocp.constraints.idxbxe_0 = np.array(idxbxe_0_filtered)
        m["lbx_0"] = export_vec(lbx_0)
        m["ubx_0"] = export_vec(ubx_0)

        args = [v[0] for v in m.values()]
        outputs = list(m.keys())
        self.mmap = Function('mmap',[self.p_global_cat,self.p_local_cat,stage.t],args,['p_global','p_local','t'],outputs)

        tgrid = ca.MX.sym("tgrid",1,self.N+1)
        P_local = [ca.MX.sym("P_local",self.p_local_cat.shape[0]) for i in range(self.N+1)]
        resv = [self.mmap(p_global=self.p_global_cat,p_local=p_local,t=t) for t,p_local in zip(ca.horzsplit(tgrid),P_local)]

        res = {}

        for k in m.keys():
            if k.endswith("_0"):
                res[k] = resv[0][k]
            elif k.endswith("_e"):
                res[k] = resv[-1][k]
            else:
                res[k] = ca.hcat([r[k] for r in resv])

        for k in m.keys():
            res[k] = ca.fmax(ca.fmin(res[k], INF),-INF)

        self.mmap_horizon = Function('mmap_horizon',[self.p_global_cat,ca.hcat(P_local),tgrid],[res[k] for k in outputs],['p_global','p_local','t'],outputs)

        for k,v in stage._param_vals.items():
            self.set_value(stage, self, k, v)

        self.ocp.parameter_values = np.zeros(stage.p.shape[0])

        res = self.mmap(p_global=self.p_global_value,p_local=self.p_local_value[:,0],t=0)
        # Set matrices
        for k, (_,is_vec) in self.m.items():
            v = np.array(res[k])
            if is_vec:
                v = v.reshape((-1))
            v[v==-inf] = -INF
            v[v==inf] = INF
            setattr(self.ocp.constraints, k, v)

        for k, v in self.args.items():
            setattr(ocp.solver_options, k, v)


        ocp.solver_options.tf = 1 if isinstance(stage._T, FreeTime) else stage._T
        ocp.solver_options.qp_solver_iter_max = 1000
        #ocp.solver_options.tol = 1e-8
        #ocp.solver_options.print_level = 15
        ocp.solver_options.shooting_nodes =np.ndarray.flatten(np.array(self.time_grid))
        # AcadosOcpOptions

        for k,v in self.acados_options.items():
            setattr(ocp.solver_options, k, v)

        if self.feasibility_problem:
            ocp.translate_to_feasibility_problem()

        # By-pass acados's heuristic to check lbx==ubx numerically
        ocp.dims.nbxe_0 = self.ocp.constraints.idxbxe_0.shape[0]

        #self.ocp_solver = AcadosOcpSolver(ocp, json_file = 'acados_ocp_' + ocp.model.name + '.json')

        AcadosOcpSolver.generate(ocp, json_file = 'acados_ocp_' + ocp.model.name + '.json')

        c_generated_code = os.path.join(os.getcwd(), "c_generated_code")

        if not os.path.exists(os.path.join(self.build_dir_abs,"build")):
            try:
                # copy acados
                ACADOS_SOURCE_DIR = os.path.dirname(os.path.realpath(__file__)) + os.sep + "external"
                if 'ACADOS_SOURCE_DIR' in os.environ:
                    ACADOS_SOURCE_DIR = os.environ['ACADOS_SOURCE_DIR']
                recursive_overwrite(ACADOS_SOURCE_DIR, os.path.join(self.build_dir_abs,"acados"))
            except:
                pass
        recursive_overwrite(os.path.dirname(os.path.realpath(__file__)) + os.sep + "interface_generation",self.build_dir_abs)
        recursive_overwrite(c_generated_code,self.build_dir_abs)

        with open(os.path.join(self.build_dir_abs,"rockit_config.h"),"w") as out:

            for i in range(self.mmap_horizon.n_out()):
                k = self.mmap_horizon.name_out(i)
                if (k.endswith("_0") or k.endswith("_e")) and self.m[k][1]:
                    out.write(f"#define MMAP_{k.upper()}_SIZE {self.mmap_horizon.numel_out(i)}\n")
                else:
                    out.write(f"#define MMAP_{k.upper()}_SIZE1 {self.mmap_horizon.size1_out(i)}\n")
                    out.write(f"#define MMAP_{k.upper()}_SIZE2 {self.mmap_horizon.size2_out(i)}\n")
                    # Same but uppercase
            out.write(f"#define ROCKIT_N {self.N}\n")
            out.write(f"#define ROCKIT_X_SIZE1 {stage.nx}\n")
            out.write(f"#define ROCKIT_X_SIZE2 {self.N+1}\n")
            out.write(f"#define ROCKIT_U_SIZE1 {stage.nu}\n")
            out.write(f"#define ROCKIT_U_SIZE2 {self.N}\n")
            out.write(f"#define ROCKIT_P_LOCAL_SIZE1 {self.p_local_cat.shape[0]}\n")
            out.write(f"#define ROCKIT_P_LOCAL_SIZE2 {self.N+1}\n")
            out.write(f"#define ROCKIT_P_GLOBAL_SIZE1 {self.p_global_cat.shape[0]}\n")
            out.write(f"#define ROCKIT_P_GLOBAL_SIZE2 {self.p_global_cat.shape[1]}\n")

        if 0:
            with open(os.path.join(self.build_dir_abs,"after_init.c.in"), "w") as after_init:
                if self.linesearch:
                    after_init.write(f"""ocp_nlp_solver_opts_set(m->nlp_config, m->nlp_opts, "globalization","merit_backtracking");\n""")

                for k,v in self.acados_options.items():
                    if isinstance(v, bool):
                        after_init.write(f"""bool {k}={int(v)};ocp_nlp_solver_opts_set(m->nlp_config, m->nlp_opts, "{k}",&{k});\n""")
                    elif isinstance(v, str):
                        after_init.write(f"""ocp_nlp_solver_opts_set(m->nlp_config, m->nlp_opts, "{k}","{v}");\n""")
                    elif isinstance(v, bool):
                        after_init.write(f"""int {k}={v};ocp_nlp_solver_opts_set(m->nlp_config, m->nlp_opts, "{k}",&{k});\n""")
        assert subprocess.run(["cmake","-S", ".","-B", "build","-DCMAKE_BUILD_TYPE=Debug", "-DMODEL_NAME="+ self.model_name, "-DBLASFEO_EXAMPLES=OFF"], cwd=self.build_dir_abs).returncode==0
        assert subprocess.run(["cmake","--build","build","--config","Debug"], cwd=self.build_dir_abs).returncode==0
        assert subprocess.run(["cmake","--install","build","--prefix","."], cwd=self.build_dir_abs).returncode==0
        
        self.acados_driver = ca.external("acados_driver", self.build_dir_abs + "/lib/libacados_driver.so")

        X0 = DM.zeros(ocp.dims.nx, self.N+1)
        U0 = DM.zeros(ocp.dims.nu, self.N)


        for var, expr in stage._initial.items():
            if depends_on(expr, stage.t):
                expr = evalf(substitute(expr,stage.t, self.control_grid_init))

            var = substitute([var],self.raw,self.optivar)[0]
            Jx, Ju, r = linear_coeffs(var,self.X, self.U)
            Jx = evalf(Jx)
            Ju = evalf(Ju)
            r = evalf(r)
            assert r.is_zero()
            check_Js(Jx)
            check_Js(Ju)
            assert Jx.nnz()==0 or Ju.nnz()==0
            expr = reshape_number(var, expr)
            is_matrix = False
            if expr.shape[1]!= var.shape[1]:
                if expr.shape[1]==self.N*var.shape[1] or expr.shape[1]==(self.N+1)*var.shape[1]:
                    is_matrix = True
                else:
                    raise Exception("Initial guess of wrong shape")
            assert expr.shape[0]==var.shape[0]
            if Jx.sparsity().get_col():
                for k in range(self.N+1):
                    X0[Jx.sparsity().get_col(),k] = expr[Jx.row(),k if is_matrix else 0]
            if Ju.sparsity().get_col():
                for k in range(self.N):
                    U0[Ju.sparsity().get_col(),k] = expr[Ju.row(),k if is_matrix else 0]

        self.X0 = np.array(X0)
        self.U0 = np.array(U0)


    def to_function(self, stage, name, args, results, *margs):

        args = [stage.value(a) for a in args]

        controls = ca.hcat(self.U_gist)
        states = ca.hcat(self.X_gist)
        p_stage = ca.hcat(self.P_stage_gist)
        p_global = self.P_global_gist

        gist_list = self.U_gist+self.X_gist+self.P_stage_gist+[self.P_global_gist,self.T_gist]
        try:
            helper0 = Function("helper0", args, gist_list, {"allow_free":True})
        except:
            helper0 = Function("helper0", args, gist_list)
        if helper0.has_free():
            helper0_free = helper0.free_mx()
            [controls, states, p_stage, p_global, T_gist] = ca.substitute([controls, states, p_stage, p_global, self.T_gist], helper0_free, self.initial_value(stage, helper0_free))

        res = self.mmap_horizon(p_local=p_stage,p_global=p_global,t=self.normalized_time_grid.T)

        res["x0"] = states
        res["u0"] = controls

        res["p_local"] = p_stage
        res["p_global"] = p_global

        res = self.acados_driver(**res)

        arg_in = ca.substitute(results,gist_list, ca.horzsplit(res["u"])+ca.horzsplit(res["x"])+ca.horzsplit(p_stage)+[p_global,T_gist])

        ret = Function(name, args, arg_in, *margs)
        assert not ret.has_free()
        return ret
    
    def solve(self, stage):
        res = self.mmap_horizon(p_global=self.p_global_value,p_local=self.p_local_value,t=self.normalized_time_grid.T)

        res["p_local"] = self.p_local_value
        res["p_global"] = self.p_global_value
    
        res["x0"] = self.X0
        res["u0"] = self.U0
        ret = self.acados_driver(**res)

        return OcpSolution(SolWrapper(self, ca.vec(ret["x"]), ca.vec(ret["u"]), 0, self.p_global_value, self.p_local_value), stage)

    def initial_value(self, stage, expr):
        # check if expr is a list
        lst = isinstance(expr, list) 
        if not lst:
            expr = [expr]
        # expr = stage.value(expr)
        [_,states] = stage.sample(stage.x,grid='control')
        [_,controls] = stage.sample(stage.u,grid='control-')
        variables = stage.value(veccat(stage.variables['']))

        helper_in = [self.P_global_gist,ca.horzcat(*self.P_stage_gist),states,controls,variables,self.T_gist]
        helper = Function("helper", helper_in, expr)
        ret = [s.toarray(simplify=True) for s in helper.call([self.p_global_value, self.p_local_value, self.X0, self.U0, 0, self.inits[0][1]])]
        if not lst :
            return ret[0]
        return ret
class SolWrapper:
    def __init__(self, method, x, u, T, P_global, P_stage):
        self.method = method
        self.x = x
        self.u = u
        self.P_global = P_global
        self.P_stage = P_stage
        self.T = T

    def value(self, expr, *args,**kwargs):
        placeholders = self.method.stage.placeholders_transcribed
        ret = evalf(substitute([placeholders(expr)],[self.method.gist],[veccat(self.x, self.u, self.P_global, self.P_stage, self.T)])[0])
        return ret.toarray(simplify=True)


from casadi import substitute, depends_on, vvcat, DM
from .casadi_helpers import HashDict
from collections import defaultdict

class TranscribedPlaceholders:
    def __init__(self):
        self.clear()

    def clear(self):
        self.pool = [HashDict() for i in range(2)]
        self.mark_dirty()

    def mark_dirty(self):
        self.is_dirty = True

    def __getitem__(self, i):
        return self.pool[i-1]

    def _replace(self,args,ks,vs):
        # No use doing placeholder substitution on DM
        if isinstance(vvcat(args), DM):
            return args

        kv = vvcat(ks)

        # Fixed-point iteration
        while depends_on(vvcat(args), kv):
            args = substitute(args, ks, vs)

        return args

    def __call__(self, args, max_phase=2, preference=None, verbose=False):
        if not isinstance(args, list):
            return self([args],max_phase=max_phase,preference=preference,verbose=verbose)[0]

        if preference is None:
            preference = ['normal','normal.normal']

        if not isinstance(args, list):
            return self.__call__([args])[0]

        def select(value):
            if len(value)==1:
                return list(value.values())[0]
            for k,v in value.items():
                if k in preference:
                    return v
            raise Exception("ambiguous")

        if max_phase==1:
            ks = list(self[1])
            vs = [select(self[1][e]) for e in ks]
        elif max_phase==2:
            ks = list(self[2])
            vs = [select(self[2][e]) for e in ks]

            k = [k for k in self[1].keys() if k not in self[2]]        
            ks += k
            vs += [select(self[1][e]) for e in k]
            kv = vvcat(ks)

        if verbose:
            print(self.pool)
            print(ks,vs)
        return self._replace(args, ks, vs)